import os
import wave
import unittest

import pytest

from io import BytesIO

from pyglet.media.synthesis import *


local_dir = os.path.dirname(__file__)
test_data_path = os.path.abspath(os.path.join(local_dir, '..', '..', 'data'))
del local_dir


def get_test_data_file(*file_parts):
    """Get a file from the test data directory in an OS independent way.

    Supply relative file name as you would in os.path.join().
    """
    return os.path.join(test_data_path, *file_parts)


classes = [Silence, WhiteNoise, Sine, Triangle, Sawtooth, Square]
sample_rates = [44800, 11025]


@pytest.mark.parametrize('source_class', classes)
@pytest.mark.parametrize('sample_rate', sample_rates)
def test_instantiation(source_class, sample_rate):
    source = source_class(duration=1, sample_rate=sample_rate)


@pytest.mark.parametrize('source_class', classes)
@pytest.mark.parametrize('sample_rate', sample_rates)
def test_total_duration(source_class, sample_rate):
    source = source_class(duration=1)                       # One second of audio...
    expected_bytes = source.audio_format.bytes_per_second   # should match this.
    audio_data = source.get_audio_data(expected_bytes + 100)

    assert audio_data.length == pytest.approx(expected_bytes)
    assert audio_data.duration == pytest.approx(1.0)
    assert len(audio_data.data) == pytest.approx(expected_bytes)

    # Should now be out of data
    last_data = source.get_audio_data(100)
    assert last_data is None


@pytest.mark.parametrize('source_class', classes)
@pytest.mark.parametrize('sample_rate', sample_rates)
def test_generated_bytes(source_class, sample_rate):
    if source_class == WhiteNoise:
        return

    source = source_class(duration=1, sample_rate=sample_rate)
    source_name = source_class.__name__.lower()
    filename = "synthesis_{0}_{1}_{2}_1ch.wav".format(source_name, 16, sample_rate)

    with wave.open(get_test_data_file('media', filename)) as f:
        loaded_bytes = f.readframes(-1)
        generated_data = source.get_audio_data(source._max_offset)
        bytes_buffer = BytesIO(generated_data.data).getvalue()
        # Compare a small chunk, to avoid hanging on mismatch:
        assert bytes_buffer[:1000] == loaded_bytes[:1000], "Generated bytes do not match sample wave file."



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Literal

from pyglet import app, display, gl, window

if TYPE_CHECKING:
    from pyglet.gl import Config
    from pyglet.window import BaseWindow


class Display:
    """A display device supporting one or more screens."""

    name: str = None
    """Name of this display, if applicable."""

    x_screen: int = None
    """The X11 screen number of this display, if applicable."""

    def __init__(self, name: str = None, x_screen: int = None) -> None:
        """Create a display connection for the given name and screen.

        On X11, :attr:`name` is of the form ``"hostname:display"``, where the
        default is usually ``":1"``.  On X11, :attr:`x_screen` gives the X 
        screen number to use with this display.  A pyglet display can only be 
        used with one X screen; open multiple display connections to access
        multiple X screens.  
        
        Note that TwinView, Xinerama, xrandr and other extensions present
        multiple monitors on a single X screen; this is usually the preferred
        mechanism for working with multiple monitors under X11 and allows each
        screen to be accessed through a single pyglet`~pyglet.display.Display`

        On platforms other than X11, :attr:`name` and :attr:`x_screen` are 
        ignored; there is only a single display device on these systems.
        """
        display._displays.add(self)

    def get_screens(self) -> list[Screen]:
        """Get the available screens.

        A typical multi-monitor workstation comprises one :class:`Display`
        with multiple :class:`Screen` s.  This method returns a list of 
        screens which can be enumerated to select one for full-screen display.

        For the purposes of creating an OpenGL config, the default screen
        will suffice.

        :rtype: list of :class:`Screen`
        """
        raise NotImplementedError('abstract')

    def get_default_screen(self) -> Screen:
        """Get the default (primary) screen as specified by the user's operating system preferences."""
        screens = self.get_screens()
        for screen in screens:
            if screen.x == 0 and screen.y == 0:
                return screen

        # No Primary screen found?
        return screens[0]

    def get_windows(self) -> list[BaseWindow]:
        """Get the windows currently attached to this display."""
        return [win for win in app.windows if win.display is self]


class Screen(abc.ABC):
    """A virtual monitor that supports fullscreen windows.

    Screens typically map onto a physical display such as a
    monitor, television or projector.  Selecting a screen for a window
    has no effect unless the window is made fullscreen, in which case
    the window will fill only that particular virtual screen.

    The :attr:`width` and :attr:`height` attributes of a screen give the 
    current resolution of the screen.  The :attr:`x` and :attr:`y` attributes 
    give the global location of the top-left corner of the screen.  This is 
    useful for determining if screens are arranged above or next to one 
    another.

    Use :func:`~Display.get_screens` or :func:`~Display.get_default_screen`
    to obtain an instance of this class.
    """

    def __init__(self, display: Display, x: int, y: int, width: int, height: int) -> None:
        self.display = display
        """Display this screen belongs to."""
        self.x = x
        """Left edge of the screen on the virtual desktop."""
        self.y = y
        """Top edge of the screen on the virtual desktop."""
        self.width = width
        """Width of the screen, in pixels."""
        self.height = height
        """Height of the screen, in pixels."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.x}, y={self.y}, width={self.width}, height={self.height})"

    def get_best_config(self, template: Config = None) -> Config:
        """Get the best available GL config.

        Any required attributes can be specified in ``template``.  If
        no configuration matches the template,
        :class:`~pyglet.window.NoSuchConfigException` will be raised.
        A configuration supported by the platform that best fulfils
        the needs described by the template.

        :deprecated: Use :meth:`pyglet.gl.Config.match`.

        Args:
            template:
                A configuration with desired attributes filled in.
        """
        configs = None
        if template is None:
            for template_config in [gl.Config(double_buffer=True, depth_size=24, major_version=3, minor_version=3),
                                    gl.Config(double_buffer=True, depth_size=16, major_version=3, minor_version=3),
                                    None]:
                try:
                    configs = self.get_matching_configs(template_config)
                    break
                except window.NoSuchConfigException:
                    pass
        else:
            configs = self.get_matching_configs(template)
        if not configs:
            raise window.NoSuchConfigException()
        return configs[0]

    def get_matching_configs(self, template: Config) -> list[Config]:
        """Get a list of configs that match a specification.

        Any attributes specified in `template` will have values equal
        to or greater in each returned config.  If no configs satisfy
        the template, an empty list is returned.

        :deprecated: Use :meth:`pyglet.gl.Config.match`.

        Args:
            template:
                A configuration with desired attributes filled in.
        """
        raise NotImplementedError('abstract')

    def get_modes(self) -> list[ScreenMode]:
        """Get a list of screen modes supported by this screen.

        .. versionadded:: 1.2
        """
        raise NotImplementedError('abstract')

    def get_mode(self) -> ScreenMode:
        """Get the current display mode for this screen.

        .. versionadded:: 1.2
        """
        raise NotImplementedError('abstract')

    def get_closest_mode(self, width: int, height: int) -> ScreenMode:
        """Get the screen mode that best matches a given size.

        If no supported mode exactly equals the requested size, a larger one
        is returned; or ``None`` if no mode is large enough.

        .. versionadded:: 1.2
        """
        # Best mode is one with the smallest resolution larger than width/height,
        # with depth and refresh rate equal to current mode.
        current = self.get_mode()

        best = None
        for mode in self.get_modes():
            # Reject resolutions that are too small
            if mode.width < width or mode.height < height:
                continue

            if best is None:
                best = mode

            # Must strictly dominate dimensions
            if (mode.width <= best.width and mode.height <= best.height and
                    (mode.width < best.width or mode.height < best.height)):
                best = mode

            # Preferably match rate, then depth.
            if mode.width == best.width and mode.height == best.height:
                points = 0
                if mode.rate == current.rate:
                    points += 2
                if best.rate == current.rate:
                    points -= 2
                if mode.depth == current.depth:
                    points += 1
                if best.depth == current.depth:
                    points -= 1
                if points > 0:
                    best = mode
        return best

    def set_mode(self, mode: ScreenMode) -> None:
        """Set the display mode for this screen.

        The mode must be one previously returned by :meth:`get_mode` or
        :meth:`get_modes`.
        """
        raise NotImplementedError('abstract')

    def restore_mode(self) -> None:
        """Restore the screen mode to the user's default."""
        raise NotImplementedError('abstract')

    def get_dpi(self):
        """Get the DPI of the screen."""
        raise NotImplementedError('abstract')

    def get_scale(self) -> float:
        """Get the pixel scale ratio of the screen."""
        raise NotImplementedError('abstract')

    @abc.abstractmethod
    def get_display_id(self) -> str | int:
        """Get a unique identifier for the screen.

        .. versionadded: 2.1.8
        """

    @abc.abstractmethod
    def get_monitor_name(self) -> str | Literal["Unknown"]:
        """Get a friendly name for the screen if available.

        Windows and Mac OSX report what the system will see the screen as.

        On Linux, there is no API for retrieving the monitor name, other than manually decoding the EDID information.
        Instead, it will return the connection name.

        If no name can be queried, a default name of Unknown will be returned.

        .. versionadded: 2.1.8
        """


class ScreenMode:
    """Screen resolution and display settings.

    Applications should not construct `ScreenMode` instances themselves; see
    :meth:`Screen.get_modes`.

    The :attr:`depth` and :attr:`rate` variables may be ``None`` if the 
    operating system does not provide relevant data.

    .. versionadded:: 1.2
    """

    width: int = None
    """Width of screen, in pixels."""

    height: int = None
    """Height of screen, in pixels."""

    depth: int = None
    """Pixel color depth, in bits per pixel."""

    rate: int = None
    """Screen refresh rate in Hz."""

    def __init__(self, screen: Screen) -> None:
        self.screen = screen

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(width={self.width!r}, height={self.height!r}, depth={self.depth!r}, rate={self.rate})'


class Canvas:
    """Abstract drawing area.

    Canvases are used internally by pyglet to represent drawing areas --
    either within a window or full-screen.

    .. versionadded:: 1.2
    """

    def __init__(self, display: Display) -> None:
        self.display = display
        """Display this canvas was created on."""



"""Extensible attributed text format for representing pyglet formatted documents."""
from __future__ import annotations

import ast
import re
from typing import TYPE_CHECKING

import pyglet

if TYPE_CHECKING:
    from pyglet.resource import Location

_pattern = re.compile(r"""
    (?P<escape_hex>\{\#x(?P<escape_hex_val>[0-9a-fA-F]+)\})
  | (?P<escape_dec>\{\#(?P<escape_dec_val>[0-9]+)\})
  | (?P<escape_lbrace>\{\{)
  | (?P<escape_rbrace>\}\})
  | (?P<attr>\{
        (?P<attr_name>[^ \{\}]+)\s+
        (?P<attr_val>[^\}]+)\})
  | (?P<nl_hard1>\n(?=[ \t]))
  | (?P<nl_hard2>\{\}\n)
  | (?P<nl_soft>\n(?=\S))
  | (?P<nl_para>\n\n+)
  | (?P<text>[^\{\}\n]+)
    """, re.VERBOSE | re.DOTALL)


class AttributedTextDecoder(pyglet.text.DocumentDecoder):  # noqa: D101

    def __init__(self) -> None:  # noqa: D107
        self.doc = pyglet.text.document.FormattedDocument()
        self.length = 0
        self.attributes = {}

    def decode(self, text: str, location: Location | None = None) -> pyglet.text.document.FormattedDocument:  # noqa: ARG002
        next_trailing_space = True
        trailing_newline = True

        for m in _pattern.finditer(text):
            group = m.lastgroup
            trailing_space = True
            if group == "text":
                t = m.group("text")
                self.append(t)
                trailing_space = t.endswith(" ")
                trailing_newline = False
            elif group == "nl_soft":
                if not next_trailing_space:
                    self.append(" ")
                trailing_newline = False
            elif group in ("nl_hard1", "nl_hard2"):
                self.append("\n")
                trailing_newline = True
            elif group == "nl_para":
                self.append(m.group("nl_para")[1:])  # ignore the first \n
                trailing_newline = True
            elif group == "attr":
                value = ast.literal_eval(m.group("attr_val"))
                name = m.group("attr_name")
                if name[0] == ".":
                    if trailing_newline:
                        self.attributes[name[1:]] = value
                    else:
                        self.doc.set_paragraph_style(self.length, self.length, {name[1:]: value})
                else:
                    self.attributes[name] = value
            elif group == "escape_dec":
                self.append(chr(int(m.group("escape_dec_val"))))
            elif group == "escape_hex":
                self.append(chr(int(m.group("escape_hex_val"), 16)))
            elif group == "escape_lbrace":
                self.append("{")
            elif group == "escape_rbrace":
                self.append("}")
            next_trailing_space = trailing_space

        return self.doc

    def append(self, text: str) -> None:
        self.doc.insert_text(self.length, text, self.attributes)
        self.length += len(text)
        self.attributes.clear()



"""Drivers for playing back media."""

import sys
import atexit

import pyglet

_debug = pyglet.options['debug_media']
_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run


if _is_pyglet_doc_run:
    from . import silent
    _audio_driver = silent.create_audio_driver()

else:

    for driver_name in pyglet.options['audio']:
        try:
            if driver_name == 'pulse':
                from . import pulse

                _audio_driver = pulse.create_audio_driver()
                break
            elif driver_name == 'xaudio2':
                if pyglet.compat_platform in ('win32', 'cygwin'):
                    from pyglet.libs.win32.constants import WINDOWS_8_OR_GREATER

                    if WINDOWS_8_OR_GREATER:
                        from . import xaudio2

                        _audio_driver = xaudio2.create_audio_driver()
                        break
            elif driver_name == 'directsound':
                if pyglet.compat_platform in ('win32', 'cygwin'):
                    from . import directsound

                    _audio_driver = directsound.create_audio_driver()
                    break
            elif driver_name == 'openal':
                from . import openal

                _audio_driver = openal.create_audio_driver()
                break
            elif driver_name == 'silent':
                from . import silent

                _audio_driver = silent.create_audio_driver()
                break
        except Exception:
            if _debug:
                print(f'Error importing driver {driver_name}:')
                import traceback

                traceback.print_exc()

    else:
        from . import silent
        _audio_driver = silent.create_audio_driver()


def get_audio_driver():
    """Get the preferred audio driver for the current platform.

    See :data:`pyglet.options` ``audio``, and the Programming guide,
    section :doc:`/programming_guide/media` for more information on
    setting the preferred driver.

    Returns:
        AbstractAudioDriver : The concrete implementation of the preferred
                              audio driver for this platform.
    """
    return _audio_driver


def _delete_audio_driver():
    global _audio_driver

    from .. import Source
    for p in Source._players:
        p.delete()
    del Source._players

    _audio_driver.delete()
    _audio_driver = None


atexit.register(_delete_audio_driver)




from dataclasses import dataclass
from typing import Any, ItemsView, Literal, Sequence

from . import app as app
from . import clock as clock
from . import customtypes as customtypes
from . import display as display
from . import event as event
from . import font as font
from . import gl as gl
from . import graphics as graphics
from . import gui as gui
from . import image as image
from . import input as input
from . import lib as lib
from . import math as math
from . import media as media
from . import model as model
from . import resource as resource
from . import shapes as shapes
from . import sprite as sprite
from . import text as text
from . import window as window

version: str
MIN_PYTHON_VERSION: tuple[int, int]
MIN_PYTHON_VERSION_STR: str
compat_platform: str
env: str
value: str

@dataclass
class Options:
    audio: Sequence[str]
    debug_font: bool
    debug_gl: bool
    debug_gl_trace: bool
    debug_gl_trace_args: bool
    debug_gl_shaders: bool
    debug_graphics_batch: bool
    debug_lib: bool
    debug_media: bool
    debug_texture: bool
    debug_trace: bool
    debug_trace_args: bool
    debug_trace_depth: int
    debug_trace_flush: bool
    debug_win32: bool
    debug_input: bool
    debug_x11: bool
    debug_com: bool
    shadow_window: bool
    vsync: bool | None
    xsync: bool
    xlib_fullscreen_override_redirect: bool
    search_local_libs: bool
    win32_gdi_font: bool
    text_antialiasing: bool
    headless: bool
    headless_device: int
    text_shaping: Literal["platform", "harfbuzz", False]
    dw_legacy_naming: bool
    win32_disable_xinput: bool
    com_mta: bool
    osx_alt_loop: bool
    dpi_scaling: Literal["real", "scaled", "stretch"]
    shader_bind_management: bool

    def get(self, item: str, default: Any = None) -> Any:
        ...

    def items(self) -> ItemsView[str, Any]:
        ...

    def __getitem__(self, item: str) -> Any:
        ...

    def __setitem__(self, key: str, item: Any) -> None:
        ...

options: Options



import pyglet
import math
from pyglet.window import key
from . import physicalobject, resources


class Player(physicalobject.PhysicalObject):
    """Physical object that responds to user input"""

    def __init__(self, *args, **kwargs):
        super().__init__(img=resources.player_image, *args, **kwargs)

        # Create a child sprite to show when the ship is thrusting
        self.engine_sprite = pyglet.sprite.Sprite(img=resources.engine_image, *args, **kwargs)
        self.engine_sprite.visible = False

        # Set some easy-to-tweak constants
        self.thrust = 300.0
        self.rotate_speed = 200.0

        # Let pyglet handle keyboard events for us
        self.key_handler = key.KeyStateHandler()

    def update(self, dt):
        # Do all the normal physics stuff
        super().update(dt)

        if self.key_handler[key.LEFT]:
            self.rotation -= self.rotate_speed * dt
        if self.key_handler[key.RIGHT]:
            self.rotation += self.rotate_speed * dt

        if self.key_handler[key.UP]:
            angle_radians = -math.radians(self.rotation)
            force_x = math.cos(angle_radians) * self.thrust * dt
            force_y = math.sin(angle_radians) * self.thrust * dt
            self.velocity_x += force_x
            self.velocity_y += force_y

            # If thrusting, update the engine sprite
            self.engine_sprite.rotation = self.rotation
            self.engine_sprite.x = self.x
            self.engine_sprite.y = self.y
            self.engine_sprite.visible = True
        else:
            # Otherwise, hide it
            self.engine_sprite.visible = False

    def delete(self):
        # We have a child sprite which must be deleted when this object
        # is deleted from batches, etc.
        self.engine_sprite.delete()
        super().delete()



import unittest

from pyglet.image import *
from pyglet.window import *


def colorbyte(color):
    return bytes((color,))


class ImageGridTestCase(unittest.TestCase):
    """Test the ImageGrid for textures."""

    def set_grid_image(self, itemwidth, itemheight, rows, cols, rowpad, colpad):
        data = b''
        color = 1
        width = itemwidth * cols + colpad * (cols - 1)
        height = itemheight * rows + rowpad * (rows - 1)
        for row in range(rows):
            rowdata = b''
            for col in range(cols):
                rowdata += colorbyte(color) * itemwidth
                if col < cols - 1:
                    rowdata += b'\0' * colpad
                color += 1

            data += rowdata * itemheight
            if row < rows - 1:
                data += (width * b'\0') * rowpad
        assert len(data) == width * height
        self.image = ImageData(width, height, 'R', data)
        self.grid = ImageGrid(self.image, rows, cols,
                              itemwidth, itemheight, rowpad, colpad).get_texture_sequence()

    def check_cell(self, cellimage, cellindex):
        self.assertTrue(cellimage.width == self.grid.item_width)
        self.assertTrue(cellimage.height == self.grid.item_height)

        color = colorbyte(cellindex + 1)
        cellimage = cellimage.get_image_data()
        data = cellimage.get_data('R', cellimage.width)
        self.assertTrue(data == color * len(data))

    def setUp(self):
        self.w = Window(visible=False)

    def tearDown(self) -> None:
        self.w.close()

    def testSquare(self):
        # Test a 3x3 grid with no padding and 4x4 images
        rows = cols = 3
        self.set_grid_image(4, 4, rows, cols, 0, 0)
        for i in range(rows * cols):
            self.check_cell(self.grid[i], i)

    def testRect(self):
        # Test a 2x5 grid with no padding and 3x8 images
        rows, cols = 2, 5
        self.set_grid_image(3, 8, rows, cols, 0, 0)
        for i in range(rows * cols):
            self.check_cell(self.grid[i], i)

    def testPad(self):
        # Test a 5x3 grid with rowpad=3 and colpad=7 and 10x9 images
        rows, cols = 5, 3
        self.set_grid_image(10, 9, rows, cols, 3, 7)
        for i in range(rows * cols):
            self.check_cell(self.grid[i], i)

    def testTuple(self):
        # Test tuple access
        rows, cols = 3, 4
        self.set_grid_image(5, 5, rows, cols, 0, 0)
        for row in range(rows):
            for col in range(cols):
                self.check_cell(self.grid[(row, col)], row * cols + col)

    def testRange(self):
        # Test range access
        rows, cols = 4, 3
        self.set_grid_image(10, 1, rows, cols, 0, 0)
        images = self.grid[4:8]
        for i, image in enumerate(images):
            self.check_cell(image, i + 4)

    def testTupleRange(self):
        # Test range over tuples
        rows, cols = 10, 10
        self.set_grid_image(4, 4, rows, cols, 0, 0)
        images = self.grid[(3, 2):(6, 5)]
        i = 0
        for row in range(3, 6):
            for col in range(2, 5):
                self.check_cell(images[i], row * cols + col)
                i += 1



from ctypes import memmove, byref, c_uint32, sizeof, cast, c_void_p, create_string_buffer, POINTER, c_char, \
    c_long

from pyglet.libs.darwin import cf, CFSTR
from pyglet.libs.darwin.coreaudio import kCFURLPOSIXPathStyle, AudioStreamBasicDescription, ca, ExtAudioFileRef, \
    kExtAudioFileProperty_FileDataFormat, kAudioFormatLinearPCM, kAudioFormatFlagIsSignedInteger, \
    kAudioFormatFlagIsPacked, kExtAudioFileProperty_ClientDataFormat, AudioFile_ReadProc, \
    AudioFile_GetSizeProc, AudioBufferList, kExtAudioFileProperty_FileLengthFrames, AudioFileID, err_check
from pyglet.media import StreamingSource, StaticSource
from pyglet.media.codecs import AudioFormat, MediaDecoder, AudioData


class MemoryFileObject:

    def __init__(self, file):
        self.file = file

        if not getattr(self.file, 'seek', None) or not getattr(self.file, 'tell', None):
            raise Exception("File object does not support seeking.")

        # Seek to end of file to get the filesize.
        self.file.seek(0, 2)
        self.file_size = self.file.tell()
        self.file.seek(0)  # Put cursor back at the beginning.
        self.data = []

        def read_data_cb(ref, offset, requested_length, buffer, actual_count):
            self.file.seek(offset)
            data = self.file.read(requested_length)
            data_size = len(data)
            memmove(buffer, data, data_size)
            actual_count.contents.value = data_size  # Actual read.
            return 0

        def getsize_cb(ref):
            return self.file_size

        self.getsize_func = AudioFile_GetSizeProc(getsize_cb)
        self.read_func = AudioFile_ReadProc(read_data_cb)


class CoreAudioSource(StreamingSource):
    def __init__(self, filename, file=None):
        self._bl = None
        self._file = file
        self._deleted = False
        self._file_obj = None
        self._audfile = None
        self._audref = None

        audref = ExtAudioFileRef()
        if file is None:
            fn_str = CFSTR(filename)
            url_ref = cf.CFURLCreateWithFileSystemPath(None, fn_str, kCFURLPOSIXPathStyle, False)

            err_check(ca.ExtAudioFileOpenURL(url_ref, byref(audref)))
            cf.CFRelease(fn_str)
        else:
            self.file_obj = MemoryFileObject(file)

            self._audfile = AudioFileID()

            err_check(ca.AudioFileOpenWithCallbacks(
                None, self.file_obj.read_func, None, self.file_obj.getsize_func, None,
                0,
                byref(self._audfile))
            )

            err_check(ca.ExtAudioFileWrapAudioFileID(self._audfile, False, byref(audref)))

        self._audref = audref

        format_info = AudioStreamBasicDescription()
        size = c_uint32(sizeof(format_info))
        err_check(ca.ExtAudioFileGetProperty(self._audref,
                                             kExtAudioFileProperty_FileDataFormat,
                                             byref(size),
                                             byref(format_info)))

        self.convert_desc = self.convert_format(format_info)

        err_check(ca.ExtAudioFileSetProperty(
            self._audref,
            kExtAudioFileProperty_ClientDataFormat,
            sizeof(self.convert_desc),
            byref(self.convert_desc)
        ))

        length = c_long()
        size = c_uint32(sizeof(format_info))
        # File length.
        err_check(ca.ExtAudioFileGetProperty(
            self._audref,
            kExtAudioFileProperty_FileLengthFrames,
            byref(size),
            byref(length)
        ))

        self.audio_format = AudioFormat(channels=self.convert_desc.mChannelsPerFrame,
                                        sample_size=self.convert_desc.mBitsPerChannel,
                                        sample_rate=int(self.convert_desc.mSampleRate))

        self._num_frames = length.value
        self._bytes_per_frame = self.convert_desc.mBytesPerFrame
        self._duration = self._num_frames / self.convert_desc.mSampleRate
        self._duration_per_frame = self._duration / self._num_frames

    @staticmethod
    def convert_format(original_desc, bitdepth=16):
        adesc = AudioStreamBasicDescription()
        adesc.mSampleRate = original_desc.mSampleRate
        adesc.mFormatID = kAudioFormatLinearPCM
        adesc.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked
        adesc.mChannelsPerFrame = original_desc.mChannelsPerFrame
        adesc.mBitsPerChannel = bitdepth
        adesc.mBytesPerPacket = original_desc.mChannelsPerFrame * adesc.mBitsPerChannel // 8
        adesc.mFramesPerPacket = 1
        adesc.mBytesPerFrame = adesc.mBytesPerPacket
        return adesc

    def __del__(self):
        if self._file:
            self._file.close()
            self._file = None

        if self._audfile:
            err_check(ca.AudioFileClose(self._audfile))
            self._audfile = None

        if self._audref:
            err_check(ca.ExtAudioFileDispose(self._audref))
            self._audref = None

    def get_audio_data(self, num_bytes, compensation_time=0.0):
        num_frames = c_uint32(num_bytes // self.convert_desc.mBytesPerFrame)

        if not self._bl:
            buffer = create_string_buffer(num_bytes)
            self._bl = AudioBufferList()
            self._bl.mNumberBuffers = 1
            self._bl.mBuffers[0].mNumberChannels = self.convert_desc.mChannelsPerFrame
            self._bl.mBuffers[0].mDataByteSize = num_bytes
            self._bl.mBuffers[0].mData = cast(buffer, c_void_p)

        while True:
            ca.ExtAudioFileRead(self._audref, byref(num_frames), byref(self._bl))

            size = self._bl.mBuffers[0].mDataByteSize
            if not size:
                break

            data = cast(self._bl.mBuffers[0].mData, POINTER(c_char))
            slice = data[:size]
            return AudioData(slice, size, 0.0, size / self.audio_format.sample_rate, [])

        return None

    def seek(self, timestamp):
        self._bl = None  # invalidate buffer list.
        timestamp = max(0.0, min(timestamp, self._duration))
        position = int(timestamp / self._duration_per_frame)
        ca.ExtAudioFileSeek(self._audref, position)


#########################################
#   Decoder class:
#########################################

class CoreAudioDecoder(MediaDecoder):

    def get_file_extensions(self):
        return '.aac', '.ac3', '.aif', '.aiff', '.aifc', '.caf', '.mp3', '.mp4', '.m4a', '.snd', '.au', '.sd2', '.wav'

    def decode(self, filename, file, streaming=True):
        if streaming:
            return CoreAudioSource(filename, file)
        else:
            return StaticSource(CoreAudioSource(filename, file))


def get_decoders():
    return [CoreAudioDecoder()]


def get_encoders():
    return []



"""This is a simple example that visually confirms """

import pyglet
from pyglet.gl import (
    glClearColor,
    GL_TRIANGLES,
)


# pyglet.options['debug_gl_shaders'] = True

window = pyglet.window.Window(width=540, height=540, resizable=True)
batch = pyglet.graphics.Batch()
print(f"OpenGL Context: {window.context.get_info().version}")

program = pyglet.graphics.get_default_shader()

##########################################################
#   TESTS !
##########################################################
label = pyglet.text.Label("Scroll the mouse wheel", x=0, y=180, dpi=200, color=(255, 25, 255, 150), batch=batch)

vertex_list = program.vertex_list(3, GL_TRIANGLES,
                                  position=('f', (100, 300, 0,  200, 250, 0,  200, 350, 0)),
                                  colors=('f', (1, 0, 0, 1,  0, 1, 0, 1,  0.3, 0.3, 1, 1)))


def create_quad_vertex_list(x, y, z, width, height):
    return x, y, z, x + width, y, z, x + width, y + height, z, x, y + height, z


program.vertex_list_indexed(4, GL_TRIANGLES, [0, 1, 2, 0, 2, 3], batch, None,
                            position=('f', create_quad_vertex_list(480, 270, -11, 50, 50)),
                            colors=('f', (1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1)))

program.vertex_list_indexed(4, GL_TRIANGLES, [0, 1, 2, 0, 2, 3], batch, None,
                            position=('f', (400, 400, 0, 400+50, 400, 0, 400+50, 400+50, 0, 400, 400+50, 0)),
                            colors=('f', (1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1)))


img = pyglet.image.load("pyglet.png")
img.anchor_x = img.width // 2
img.anchor_y = img.height // 2
red = pyglet.image.SolidColorImagePattern((255, 0, 0, 255)).create_image(50, 50)
green = pyglet.image.SolidColorImagePattern((0, 255, 0, 255)).create_image(50, 50)
blue = pyglet.image.SolidColorImagePattern((0, 0, 255, 255)).create_image(50, 50)
white = pyglet.image.SolidColorImagePattern((255, 255, 255, 255)).create_image(50, 50)

sprites = [pyglet.sprite.Sprite(img=img, x=60, y=80, batch=batch),
           pyglet.sprite.Sprite(img=img, x=110, y=90, batch=batch),
           pyglet.sprite.Sprite(img=img, x=160, y=100, batch=batch),
           pyglet.sprite.Sprite(img=img, x=210, y=110, batch=batch)]
for sprite in sprites:
    sprite.opacity = 220

sprite2 = pyglet.sprite.Sprite(img=red, x=200, y=400, batch=batch)
sprite3 = pyglet.sprite.Sprite(img=green, x=300, y=300, batch=batch)
sprite4 = pyglet.sprite.Sprite(img=blue, x=400, y=200, batch=batch)
sprite5 = pyglet.sprite.Sprite(img=white, x=500, y=100, batch=batch)

standalone_sprite = pyglet.sprite.Sprite(img=white, x=600, y=0)


##########################################################
# Modify the sprite scale value by scrolling the mouse
##########################################################

@window.event
def on_mouse_scroll(x, y, mouse, direction):
    for spr in sprites:
        spr.scale += direction / 10


###########################################################
#
###########################################################
@window.event
def on_draw():
    window.clear()
    vertex_list.draw(GL_TRIANGLES)
    batch.draw()
    standalone_sprite.draw()


def update(dt):
    for spr in sprites:
        spr.rotation += 100 * dt % 360


if __name__ == "__main__":
    glClearColor(0.2, 0.3, 0.3, 1)
    pyglet.clock.schedule_interval(update, 1/60)
    pyglet.app.run()



"""Loading of 3D scenes and models.

The :py:mod:`~pyglet.model` module provides an interface for loading 3D "scenes"
and models. A :py:class:`~pyglet.model.Scene` is a logical container that can
contain the data of one or more models, and is closely based on the design
of the glTF format.

The following example loads a ``"teapot.obj"`` file. The wavefront format
only contains a single model (mesh)::

    import pyglet

    window = pyglet.window.Window()
    batch = pyglet.graphics.Batch()

    scene = pyglet.model.load('teapot.obj')
    models = scene.create_models(batch=batch)

    @window.event
    def on_draw():
        batch.draw()

    pyglet.app.run()


You can also load scenes with :py:meth:`~pyglet.resource.scene`.
See :py:mod:`~pyglet.resource` for more information.
"""
from __future__ import annotations

from math import pi, sin, cos
from typing import TYPE_CHECKING

import pyglet
from pyglet import gl, graphics
from pyglet.math import Mat4

from .codecs import add_default_codecs as _add_default_codecs
from .codecs import registry as _codec_registry
from .codecs.base import Material, Scene, SimpleMaterial

if TYPE_CHECKING:
    from typing import BinaryIO, TextIO
    from pyglet.image import Texture
    from pyglet.graphics import Batch, Group
    from pyglet.graphics.shader import ShaderProgram
    from pyglet.graphics.vertexdomain import VertexList
    from pyglet.image import Texture
    from pyglet.model.codecs import ModelDecoder


def load(filename: str, file: BinaryIO | TextIO | None = None, decoder: ModelDecoder | None = None) -> Scene:
    """Load a 3D scene from a file.

    Args:
        filename:
            Used to guess the scene format, or to load the file if ``file`` is unspecified.
        file:
            An open file containing the source of the scene data in any supported format.
        decoder:
            The specific decoder to use to load the Scene. If None, use default decoders
            that match the filename extension.
    """
    if decoder:
        return decoder.decode(filename, file)

    return _codec_registry.decode(filename, file)


def get_default_shader() -> ShaderProgram:
    return pyglet.gl.current_context.create_program((MaterialGroup.default_vert_src, 'vertex'),
                                                    (MaterialGroup.default_frag_src, 'fragment'))


def get_default_textured_shader() -> ShaderProgram:
    return pyglet.gl.current_context.create_program((TexturedMaterialGroup.default_vert_src, 'vertex'),
                                                    (TexturedMaterialGroup.default_frag_src, 'fragment'))


class Model:
    """Instance of a 3D object.

    See the module documentation for usage.
    """

    def __init__(self, vertex_lists: list[VertexList], groups: list[Group], batch: Batch | None = None) -> None:
        """Create a model instance.

        Args:
            vertex_lists:
                A list of :py:class:`~pyglet.graphics.VertexList` or
                :py:class:`~pyglet.graphics.IndexedVertexList`.
            groups:
                A list of :py:class:`~pyglet.model.TexturedMaterialGroup`, or
                 :py:class:`~pyglet.model.MaterialGroup`. Each group corresponds
                 to a vertex list in ``vertex_lists`` at the same index.
            batch:
                The batch to add the model to. If no batch is provided,
                the model will maintain its own internal batch.
        """
        self.vertex_lists = vertex_lists
        self.groups = groups
        self._batch = batch or graphics.Batch()
        self._modelview_matrix = Mat4()

    @property
    def batch(self) -> Batch:
        """The graphics Batch that the Model belongs to.

        The Model can be migrated from one batch to another, or removed from
        a batch (for individual drawing). If not part of any batch, the Model
        will keep its own internal batch. Note that batch migration can be
        an expensive operation.
        """
        return self._batch

    @batch.setter
    def batch(self, batch: Batch | None):
        if self._batch == batch:
            return

        if batch is None:
            batch = graphics.Batch()

        for group, vlist in zip(self.groups, self.vertex_lists):
            self._batch.migrate(vlist, gl.GL_TRIANGLES, group, batch)

        self._batch = batch

    @property
    def matrix(self) -> Mat4:
        return self._modelview_matrix

    @matrix.setter
    def matrix(self, matrix: Mat4):
        self._modelview_matrix = matrix
        for group in self.groups:
            group.matrix = matrix


class BaseMaterialGroup(graphics.Group):
    default_vert_src: str
    default_frag_src: str
    matrix: Mat4 = Mat4()

    def __init__(self, material: SimpleMaterial, program: ShaderProgram, order: int = 0, parent: Group | None = None) -> None:
        super().__init__(order, parent)
        self.material = material
        self.program = program


class TexturedMaterialGroup(BaseMaterialGroup):
    default_vert_src = """#version 330 core
    in vec3 POSITION;
    in vec3 NORMAL;
    in vec2 TEXCOORD_0;
    in vec4 COLOR_0;

    out vec3 position;
    out vec3 normal;
    out vec2 texcoord_0;
    out vec4 color_0;    

    uniform WindowBlock
    {
        mat4 projection;
        mat4 view;
    } window;

    uniform mat4 model;

    void main()
    {
        mat4 mv = window.view * model;
        vec4 pos = mv * vec4(POSITION, 1.0);
        gl_Position = window.projection * pos;
        mat3 normal_matrix = transpose(inverse(mat3(mv)));

        position = pos.xyz;
        normal = normal_matrix * NORMAL;
        texcoord_0 = TEXCOORD_0;
        color_0 = COLOR_0;
    }
    """
    default_frag_src = """#version 330 core
    in vec4 color_0;
    in vec3 normal;
    in vec2 texcoord_0;
    in vec3 position;
    out vec4 final_colors;

    uniform sampler2D our_texture;

    void main()
    {
        float l = dot(normalize(-position), normalize(normal));
        vec4 tex_color = texture(our_texture, texcoord_0) * color_0;
        // 75/25 light ambient
        final_colors = tex_color * l * 0.75 + tex_color * vec4(0.25);
    }
    """

    def __init__(self, material: SimpleMaterial, program: ShaderProgram,
                 texture: Texture, order: int = 0, parent: Group | None = None):
        super().__init__(material, program, order, parent)
        self.texture = texture

    def set_state(self) -> None:
        gl.glActiveTexture(gl.GL_TEXTURE0)
        gl.glBindTexture(self.texture.target, self.texture.id)
        self.program.use()
        self.program['model'] = self.matrix

    def __hash__(self) -> int:
        return hash((self.texture.target, self.texture.id, self.program, self.order, self.parent))

    def __eq__(self, other) -> bool:
        return (self.__class__ is other.__class__ and
                self.material == other.material and
                self.texture.target == other.texture.target and
                self.texture.id == other.texture.id and
                self.program == other.program and
                self.order == other.order and
                self.parent == other.parent)


class MaterialGroup(BaseMaterialGroup):
    default_vert_src = """#version 330 core
    in vec3 POSITION;
    in vec3 NORMAL;
    in vec4 COLOR_0;

    out vec4 color_0;
    out vec3 normal;
    out vec3 position;

    uniform WindowBlock
    {
        mat4 projection;
        mat4 view;
    } window;

    uniform mat4 model;

    void main()
    {
        mat4 mv = window.view * model;
        vec4 pos = mv * vec4(POSITION, 1.0);
        gl_Position = window.projection * pos;
        mat3 normal_matrix = transpose(inverse(mat3(mv)));

        position = pos.xyz;
        color_0 = COLOR_0;
        normal = normal_matrix * NORMAL;
    }
    """
    default_frag_src = """#version 330 core
    in vec4 color_0;
    in vec3 normal;
    in vec3 position;
    out vec4 final_colors;

    void main()
    {
        float l = dot(normalize(-position), normalize(normal));
        // 75/25 light ambient
        final_colors = color_0 * l * 0.75 + color_0 * vec4(0.25);
    }
    """

    def set_state(self) -> None:
        self.program.use()
        self.program['model'] = self.matrix

    def __hash__(self):
        return hash((self.material, self.program, self.order, self.parent))

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
                self.material == other.material and
                self.program == other.program and
                self.order == other.order and
                self.parent == other.parent)


class Cube(Model):

    def __init__(self, width=1.0, height=1.0, depth=1.0, color=(1.0, 1.0, 1.0, 1.0),
                 material=None, batch=None, group=None, program=None):
        self._width = width
        self._height = height
        self._depth = depth
        self._color = color

        self._batch = batch
        self._program = program if program else get_default_shader()

        # Create a Material and Group for the Model
        self._material = material if material else SimpleMaterial(name="cube")
        self._group = pyglet.model.MaterialGroup(material=self._material, program=self._program, parent=group)

        self._vlist = self._create_vertexlist()

        super().__init__([self._vlist], [self._group], self._batch)

    def _create_vertexlist(self):
        w = self._width / 2
        h = self._height / 2
        d = self._depth / 2

        vertices = [
            -w, -h, -d,   # front, bottom-left    0
            w, -h, -d,    # front, bottom-right   1
            w, h, -d,     # front, top-right      2         Front
            -w, h, -d,    # front, top-left       3

            w, -h, d,     # back, bottom-right    4
            -w, -h, d,    # back, bottom-left     5
            -w, h, d,     # back, top-left        6         Back
            w, h, d,      # back, top-right       7

            w, -h, -d,    # front, bottom-right   8
            w, -h, d,     # back, bottom-right    9
            w, h, d,      # back, top-right      10         Right
            w, h, -d,     # front, top-right     11

            -w, -h, d,    # back, bottom-left    12
            -w, -h, -d,   # front, bottom-left   13
            -w, h, -d,    # front, top-left      14         Left
            -w, h, d,     # back, top-left       15

            -w, h, -d,    # front, top-left      16
            w, h, -d,     # front, top-right     17
            w, h, d,      # back, top-right      18         Top
            -w, h, d,     # back, top-left       19

            -w, -h, d,    # back, bottom-left    20
            w, -h, d,     # back, bottom-right   21
            w, -h, -d,    # front, bottom-right  22         Bottom
            -w, -h, -d,   # front, bottom-left   23
        ]

        normals = [0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1,     # front face
                   0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1,         # back face
                   1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0,         # right face
                   -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0,     # left face
                   0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0,         # top face
                   0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0]     # bottom face

        indices = [23, 22, 20, 22, 21, 20,  # bottom
                   19, 18, 16, 18, 17, 16,  # top
                   15, 14, 12, 14, 13, 12,  # left
                   11, 10, 8, 10, 9, 8,     # right
                   7, 6, 4, 6, 5, 4,        # back
                   3, 2, 0, 2, 1, 0]        # front

        return self._program.vertex_list_indexed(len(vertices) // 3, pyglet.gl.GL_TRIANGLES, indices,
                                                 batch=self._batch, group=self._group,
                                                 POSITION=('f', vertices),
                                                 NORMAL=('f', normals),
                                                 COLOR_0=('f', self._color * (len(vertices) // 3)))


class Sphere(Model):

    def __init__(self, radius=1.0, stacks=30, sectors=30, color=(1.0, 1.0, 1.0, 1.0),
                 material=None, batch=None, group=None, program=None):
        self._radius = radius
        self._stacks = stacks
        self._sectors = sectors
        self._color = color

        self._batch = batch
        self._program = program if program else get_default_shader()

        # Create a Material and Group for the Model
        self._material = material if material else SimpleMaterial(name="sphere")
        self._group = pyglet.model.MaterialGroup(material=self._material, program=self._program, parent=group)

        self._vlist = self._create_vertexlist()

        super().__init__([self._vlist], [self._group], self._batch)

    def _create_vertexlist(self):
        radius = self._radius / 2
        sectors = self._sectors
        stacks = self._stacks

        vertices = []
        normals = []
        indices = []

        sector_step = 2 * pi / sectors
        stack_step = pi / stacks

        for i in range(stacks + 1):
            stack_angle = pi / 2 - i * stack_step
            for j in range(sectors + 1):
                sector_angle = j * sector_step
                vertices.append(radius * cos(stack_angle) * cos(sector_angle))    # x
                vertices.append(radius * cos(stack_angle) * sin(sector_angle))    # y
                vertices.append(radius * sin(stack_angle))                             # z
                normals.append(cos(stack_angle) * cos(sector_angle))              # x
                normals.append(cos(stack_angle) * sin(sector_angle))              # y
                normals.append(sin(stack_angle))                                       # z

        # Generate indices
        for i in range(stacks):
            for j in range(sectors):
                first = i * (sectors + 1) + j
                second = first + sectors + 1
                indices.extend([first, second, second + 1])
                indices.extend([first, second + 1, first + 1])

        return self._program.vertex_list_indexed(len(vertices) // 3, pyglet.gl.GL_TRIANGLES, indices,
                                                 batch=self._batch, group=self._group,
                                                 POSITION=('f', vertices),
                                                 NORMAL=('f', normals),
                                                 COLOR_0=('f', self._color * (len(vertices) // 3)))


_add_default_codecs()



#!/usr/bin/env python
"""Bounces balls around a window and plays noises.

This is a simple demonstration of how pyglet efficiently manages many sound
channels without intervention.
"""

import random
import sys

import pyglet
from pyglet.window import key

BALL_IMAGE = 'ball.png'
BALL_SOUND = 'ball.wav'

if len(sys.argv) > 1:
    BALL_SOUND = sys.argv[1]

window = pyglet.window.Window(640, 480)

sound = pyglet.resource.media(BALL_SOUND, streaming=False)
balls_batch = pyglet.graphics.Batch()
balls = []
label = pyglet.text.Label('Press space to add a ball, backspace to remove',
                          font_size=14,
                          x=window.width // 2, y=10,
                          anchor_x='center')


class Ball(pyglet.sprite.Sprite):
    ball_image = pyglet.resource.image(BALL_IMAGE)
    width = ball_image.width
    height = ball_image.height

    def __init__(self):
        x = random.random() * (window.width - self.width)
        y = random.random() * (window.height - self.height)

        super().__init__(self.ball_image, x, y, batch=balls_batch)

        self.dx = (random.random() - 0.5) * 1000
        self.dy = (random.random() - 0.5) * 1000

    def update_position(self, dt):
        if self.x <= 0 or self.x + self.width >= window.width:
            self.dx *= -1
            sound.play()
        if self.y <= 0 or self.y + self.height >= window.height:
            self.dy *= -1
            sound.play()
        self.x += self.dx * dt
        self.y += self.dy * dt

        self.x = min(max(self.x, 0), window.width - self.width)
        self.y = min(max(self.y, 0), window.height - self.height)


@window.event
def on_key_press(symbol, modifiers):
    if symbol == key.SPACE:
        balls.append(Ball())
    elif symbol == key.BACKSPACE:
        if balls:
            del balls[-1]
    elif symbol == key.ESCAPE:
        window.has_exit = True


@window.event
def on_draw():
    window.clear()
    balls_batch.draw()
    label.draw()


def update(dt):
    for ball in balls:
        ball.update_position(dt)


if __name__ == '__main__':
    pyglet.clock.schedule_interval(update, 1 / 30.)
    pyglet.app.run()



import io
import os

import pytest
import pyglet

from pyglet.util import DecodeException

test_data_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', 'data'))


def get_test_data_file(path, file_name):
    """Get a file from the test data directory in an OS independent way.

    Supply relative file name as you would in os.path.join().
    """
    return os.path.join(test_data_path, path, file_name)


def test_load_from_disk():
    file_path = get_test_data_file('models', 'logo3d.obj')
    scene = pyglet.model.load(file_path)
    assert isinstance(scene, pyglet.model.Scene)


def test_load_from_object_str():
    file_path = get_test_data_file('models', 'logo3d.obj')
    with open(file_path, 'r') as f:
        file_obj = io.StringIO(f.read())
    scene = pyglet.model.load(file_path, file=file_obj)
    assert isinstance(scene, pyglet.model.Scene)


def test_load_from_object_bytes():
    file_path = get_test_data_file('models', 'logo3d.obj')
    with open(file_path, 'rb') as f:
        file_obj = io.BytesIO(f.read())
    scene = pyglet.model.load(file_path, file=file_obj)
    assert isinstance(scene, pyglet.model.Scene)


def test_no_decoders_available():
    # This is NOT a valid model file:
    file_path = get_test_data_file('media', 'alert.wav')

    with pytest.raises(DecodeException) as e:
        scene = pyglet.model.load(file_path)


def test_resource_module():
    folder_path = os.path.join(test_data_path, 'models')
    pyglet.resource.path.append(folder_path)
    pyglet.resource.reindex()

    scene = pyglet.resource.scene('logo3d.obj')
    assert isinstance(scene, pyglet.model.Scene)



import unittest

from pyglet.image import *
from pyglet.window import *


def colorbyte(color):
    return bytes((color,))


class TestTexture3D(unittest.TestCase):
    """Test the Texture3D for image grids."""

    def create_image(self, width, height, color):
        data = colorbyte(color) * (width * height)
        return ImageData(width, height, 'R', data)

    def check_image(self, image, width, height, color):
        self.assertTrue(image.width == width)
        self.assertTrue(image.height == height)
        image = image.get_image_data()
        data = image.get_data('R', image.width)
        self.assertTrue(data == colorbyte(color) * len(data))

    def set_grid_image(self, itemwidth, itemheight, rows, cols, rowpad, colpad):
        data = b''
        color = 1
        width = itemwidth * cols + colpad * (cols - 1)
        height = itemheight * rows + rowpad * (rows - 1)
        for row in range(rows):
            rowdata = b''
            for col in range(cols):
                rowdata += colorbyte(color) * itemwidth
                if col < cols - 1:
                    rowdata += b'\0' * colpad
                color += 1

            data += rowdata * itemheight
            if row < rows - 1:
                data += (width * b'\0') * rowpad
        assert len(data) == width * height
        self.image = ImageData(width, height, 'R', data)
        grid = ImageGrid(self.image, rows, cols,
                         itemwidth, itemheight, rowpad, colpad)
        self.grid = Texture3D.create_for_image_grid(grid)

    def check_cell(self, cellimage, cellindex):
        self.assertTrue(cellimage.width == self.grid.item_width)
        self.assertTrue(cellimage.height == self.grid.item_height)

        cellimage = cellimage.get_image_data()
        data = cellimage.get_data('R', cellimage.width)
        self.assertTrue(data == colorbyte(cellindex + 1) * len(data))

    def setUp(self):
        self.w = Window(visible=False)

    def tearDown(self) -> None:
        self.w.close()

    def test2(self):
        # Test 2 images of 32x32
        images = [self.create_image(32, 32, i + 1) for i in range(2)]
        texture = Texture3D.create_for_images(images)
        self.assertTrue(len(texture) == 2)
        for i in range(2):
            self.check_image(texture[i], 32, 32, i + 1)

    def test5(self):
        # test 5 images of 31x94  (power2 issues)
        images = [self.create_image(31, 94, i + 1) for i in range(5)]
        texture = Texture3D.create_for_images(images)
        self.assertTrue(len(texture) == 5)
        for i in range(5):
            self.check_image(texture[i], 31, 94, i + 1)

    def testSet(self):
        # test replacing an image
        images = [self.create_image(32, 32, i + 1) for i in range(3)]
        texture = Texture3D.create_for_images(images)
        self.assertTrue(len(texture) == 3)
        for i in range(3):
            self.check_image(texture[i], 32, 32, i + 1)
        texture[1] = self.create_image(32, 32, 87)
        self.check_image(texture[0], 32, 32, 1)
        self.check_image(texture[1], 32, 32, 87)
        self.check_image(texture[2], 32, 32, 3)

    def testSquare(self):
        # Test a 3x3 grid with no padding and 4x4 images
        rows = cols = 3
        self.set_grid_image(4, 4, rows, cols, 0, 0)
        for i in range(rows * cols):
            self.check_cell(self.grid[i], i)

    def testRect(self):
        # Test a 2x5 grid with no padding and 3x8 images
        rows, cols = 2, 5
        self.set_grid_image(3, 8, rows, cols, 0, 0)
        for i in range(rows * cols):
            self.check_cell(self.grid[i], i)

    def testPad(self):
        # Test a 5x3 grid with rowpad=3 and colpad=7 and 10x9 images
        rows, cols = 5, 3
        self.set_grid_image(10, 9, rows, cols, 3, 7)
        for i in range(rows * cols):
            self.check_cell(self.grid[i], i)



"""In some cases, you may want to restrict drawing to a specific part
of the Window. To accomplish this, you can use an OpenGL Scissor
area. In this example we define a custom Group that enforces this.
Any Sprites/Labels that are assigned to this Group will not be drawn
outside the specified area. Drag the mouse to move the position.
"""

import pyglet

from pyglet.sprite import Sprite
from pyglet.gl import glEnable, glScissor, glDisable, GL_SCISSOR_TEST


window = pyglet.window.Window(width=500, height=500)
batch = pyglet.graphics.Batch()

label = pyglet.text.Label("Drag the mouse to move the scissor area.", x=5, y=5, batch=batch)


@window.event
def on_draw():
    window.clear()
    batch.draw()


###################################################
# A custom Group class that enforces a Scissor area
###################################################

class ScissorGroup(pyglet.graphics.Group):
    """A Custom Group that defines a "Scissor" area.

    If a Sprite/Label is in this Group, any parts of it that
    fall outside the specified area will not be drawn.
    NOTE: You should use the same exact group instance
    for every object that will use the group, equal groups
    will still be kept separate.

    :Parameters:
        `x` : int
            The X coordinate of the Scissor area.
        `x` : int
            The X coordinate of the Scissor area.
        `width` : int
            The width of the Scissor area.
        `height` : int
            The height of the Scissor area.

    """

    def __init__(self, x, y, width, height, order=0, parent=None):
        super().__init__(order, parent)
        self.x, self.y = x, y
        self.width, self.height = width, height

    @property
    def area(self):
        return self.x, self.y, self.width, self.height

    @area.setter
    def area(self, area):
        self.x, self.y, self.width, self.height = area

    def set_state(self):
        glEnable(GL_SCISSOR_TEST)
        glScissor(int(self.x), int(self.y), self.width, self.height)

    def unset_state(self):
        glDisable(GL_SCISSOR_TEST)


###################################################
# Create an instance of our Group, and some Sprites
###################################################

# Create an instance of the ScissorGroup that defines the center of the window:
scissor_group = ScissorGroup(x=50, y=50, width=300, height=300)

# Create a bunch of Sprites, and assign them to our custom Group. Any parts
# of these Sprites that are outside the specified area will not be drawn.
sprites = []
img = pyglet.resource.image('pyglet.png')
for x in range(5):
    for y in range(5):
        sprite = Sprite(img, x*img.width, y*img.height, group=scissor_group, batch=batch)
        sprites.append(sprite)


@window.event
def on_mouse_drag(x, y, dx, dy, *etc):
    scissor_group.x += dx
    scissor_group.y += dy


pyglet.app.run()



"""Macintosh key scancode constants.

For a quick visual reference, see Inside Macintosh's keycode diagram:
http://boredzo.org/blog/wp-content/uploads/2007/05/imtx-virtual-keycodes.png

For clarity, a few constants use custom pyglet-specific names.
These include:
================ ======================
New pyglet name  Name in referenced .h
================ ======================
QZ_LCOMMAND      QZ_LMETA
QZ_LOPTION       QZ_LALT
QZ_ROPTION       QZ_RALT
QZ_RCOMMAND      QZ_RMETA

This module was originally created by consulting a 2006 version of
SDL 1.X's SDL_Quartzkeys.h. Although some names are pyglet-specific,
all values are the same as those in Mac OS X 10.6's Events.h.

To learn more, please see:

* SDL 1.2's SDL_Quartzkeys.h https://github.com/libsdl-org/SDL-1.2/blob/main/src/video/quartz/SDL_QuartzKeys.h
* Event.h from Mac OS X 10.6 https://github.com/phracker/MacOSX-SDKs/blob/master/MacOSX10.6.sdk/System/Library/Frameworks/Carbon.framework/Versions/A/Frameworks/HIToolbox.framework/Versions/A/Headers/Events.h

"""
from pyglet.window import key

# These values and most of their names are taken from SDL 1.X's
# SQL_Quartzkeys.h. Changed names are listed in the top-level
# docstring and marked with comments below.
QZ_ESCAPE = 0x35
QZ_F1 = 0x7A
QZ_F2 = 0x78
QZ_F3 = 0x63
QZ_F4 = 0x76
QZ_F5 = 0x60
QZ_F6 = 0x61
QZ_F7 = 0x62
QZ_F8 = 0x64
QZ_F9 = 0x65
QZ_F10 = 0x6D
QZ_F11 = 0x67
QZ_F12 = 0x6F
QZ_F13 = 0x69
QZ_F14 = 0x6B
QZ_F15 = 0x71
QZ_F16 = 0x6A
QZ_F17 = 0x40
QZ_F18 = 0x4F
QZ_F19 = 0x50
QZ_F20 = 0x5A
QZ_BACKQUOTE = 0x32
QZ_1 = 0x12
QZ_2 = 0x13
QZ_3 = 0x14
QZ_4 = 0x15
QZ_5 = 0x17
QZ_6 = 0x16
QZ_7 = 0x1A
QZ_8 = 0x1C
QZ_9 = 0x19
QZ_0 = 0x1D
QZ_MINUS = 0x1B
QZ_EQUALS = 0x18
QZ_BACKSPACE = 0x33
QZ_INSERT = 0x72
QZ_HOME = 0x73
QZ_PAGEUP = 0x74
QZ_NUMLOCK = 0x47
QZ_KP_EQUALS = 0x51
QZ_KP_DIVIDE = 0x4B
QZ_KP_MULTIPLY = 0x43
QZ_TAB = 0x30
QZ_q = 0x0C
QZ_w = 0x0D
QZ_e = 0x0E
QZ_r = 0x0F
QZ_t = 0x11
QZ_y = 0x10
QZ_u = 0x20
QZ_i = 0x22
QZ_o = 0x1F
QZ_p = 0x23
QZ_LEFTBRACKET = 0x21
QZ_RIGHTBRACKET = 0x1E
QZ_BACKSLASH = 0x2A
QZ_DELETE = 0x75
QZ_END = 0x77
QZ_PAGEDOWN = 0x79
QZ_KP7 = 0x59
QZ_KP8 = 0x5B
QZ_KP9 = 0x5C
QZ_KP_MINUS = 0x4E
QZ_CAPSLOCK = 0x39
QZ_a = 0x00
QZ_s = 0x01
QZ_d = 0x02
QZ_f = 0x03
QZ_g = 0x05
QZ_h = 0x04
QZ_j = 0x26
QZ_k = 0x28
QZ_l = 0x25
QZ_SEMICOLON = 0x29
QZ_QUOTE = 0x27
QZ_RETURN = 0x24
QZ_KP4 = 0x56
QZ_KP5 = 0x57
QZ_KP6 = 0x58
QZ_KP_PLUS = 0x45
QZ_LSHIFT = 0x38
QZ_z = 0x06
QZ_x = 0x07
QZ_c = 0x08
QZ_v = 0x09
QZ_b = 0x0B
QZ_n = 0x2D
QZ_m = 0x2E
QZ_COMMA = 0x2B
QZ_PERIOD = 0x2F
QZ_SLASH = 0x2C
QZ_RSHIFT = 0x3C
QZ_UP = 0x7E
QZ_KP1 = 0x53
QZ_KP2 = 0x54
QZ_KP3 = 0x55
QZ_KP_ENTER = 0x4C
QZ_LCTRL = 0x3B
QZ_LOPTION = 0x3A  # Originally QZ_LMETA
QZ_LCOMMAND = 0x37  # Originally QZ_LALT
QZ_SPACE = 0x31
QZ_RCOMMAND = 0x36  # Originally QZ_RALT
QZ_ROPTION = 0x3D  # Originally QZ_RMETA
QZ_RCTRL = 0x3E
QZ_FUNCTION = 0x3F
QZ_LEFT = 0x7B
QZ_DOWN = 0x7D
QZ_RIGHT = 0x7C
QZ_KP0 = 0x52
QZ_KP_PERIOD = 0x41

# This map contains only keys that can be directly translated independent of
# keyboard layout and locale
keymap = {
    QZ_ESCAPE: key.ESCAPE,
    QZ_F1: key.F1,
    QZ_F2: key.F2,
    QZ_F3: key.F3,
    QZ_F4: key.F4,
    QZ_F5: key.F5,
    QZ_F6: key.F6,
    QZ_F7: key.F7,
    QZ_F8: key.F8,
    QZ_F9: key.F9,
    QZ_F10: key.F10,
    QZ_F11: key.F11,
    QZ_F12: key.F12,
    QZ_F13: key.F13,
    QZ_F14: key.F14,
    QZ_F15: key.F15,
    QZ_F16: key.F16,
    QZ_F17: key.F17,
    QZ_F18: key.F18,
    QZ_F19: key.F19,
    QZ_F20: key.F20,
    QZ_BACKSPACE: key.BACKSPACE,
    QZ_INSERT: key.INSERT,
    QZ_HOME: key.HOME,
    QZ_PAGEUP: key.PAGEUP,
    QZ_NUMLOCK: key.NUMLOCK,
    QZ_KP_EQUALS: key.NUM_EQUAL,
    QZ_KP_DIVIDE: key.NUM_DIVIDE,
    QZ_KP_MULTIPLY: key.NUM_MULTIPLY,
    QZ_TAB: key.TAB,
    QZ_BACKSLASH: key.BACKSLASH,
    QZ_DELETE: key.DELETE,
    QZ_END: key.END,
    QZ_PAGEDOWN: key.PAGEDOWN,
    QZ_KP7: key.NUM_7,
    QZ_KP8: key.NUM_8,
    QZ_KP9: key.NUM_9,
    QZ_KP_MINUS: key.NUM_SUBTRACT,
    QZ_CAPSLOCK: key.CAPSLOCK,
    QZ_RETURN: key.RETURN,
    QZ_KP4: key.NUM_4,
    QZ_KP5: key.NUM_5,
    QZ_KP6: key.NUM_6,
    QZ_KP_PLUS: key.NUM_ADD,
    QZ_LSHIFT: key.LSHIFT,
    QZ_RSHIFT: key.RSHIFT,
    QZ_UP: key.UP,
    QZ_KP1: key.NUM_1,
    QZ_KP2: key.NUM_2,
    QZ_KP3: key.NUM_3,
    QZ_KP_ENTER: key.NUM_ENTER,
    QZ_LCTRL: key.LCTRL,
    QZ_LOPTION: key.LOPTION,
    QZ_LCOMMAND: key.LCOMMAND,
    QZ_SPACE: key.SPACE,
    QZ_RCOMMAND: key.RCOMMAND,
    QZ_ROPTION: key.ROPTION,
    QZ_RCTRL: key.RCTRL,
    QZ_FUNCTION: key.FUNCTION,
    QZ_LEFT: key.LEFT,
    QZ_DOWN: key.DOWN,
    QZ_RIGHT: key.RIGHT,
    QZ_KP0: key.NUM_0,
    QZ_KP_PERIOD: key.NUM_DECIMAL,
}

charmap = {
    ' ': key.SPACE,
    '!': key.EXCLAMATION,
    '"': key.DOUBLEQUOTE,
    '#': key.HASH,
    '#': key.POUND,
    '$': key.DOLLAR,
    '%': key.PERCENT,
    '&': key.AMPERSAND,
    "'": key.APOSTROPHE,
    '(': key.PARENLEFT,
    ')': key.PARENRIGHT,
    '*': key.ASTERISK,
    '+': key.PLUS,
    ',': key.COMMA,
    '-': key.MINUS,
    '.': key.PERIOD,
    '/': key.SLASH,
    '0': key._0,
    '1': key._1,
    '2': key._2,
    '3': key._3,
    '4': key._4,
    '5': key._5,
    '6': key._6,
    '7': key._7,
    '8': key._8,
    '9': key._9,
    ':': key.COLON,
    ';': key.SEMICOLON,
    '<': key.LESS,
    '=': key.EQUAL,
    '>': key.GREATER,
    '?': key.QUESTION,
    '@': key.AT,
    '[': key.BRACKETLEFT,
    '\\': key.BACKSLASH,
    ']': key.BRACKETRIGHT,
    '^': key.ASCIICIRCUM,
    '_': key.UNDERSCORE,
    '`': key.GRAVE,
    '`': key.QUOTELEFT,
    'A': key.A,
    'B': key.B,
    'C': key.C,
    'D': key.D,
    'E': key.E,
    'F': key.F,
    'G': key.G,
    'H': key.H,
    'I': key.I,
    'J': key.J,
    'K': key.K,
    'L': key.L,
    'M': key.M,
    'N': key.N,
    'O': key.O,
    'P': key.P,
    'Q': key.Q,
    'R': key.R,
    'S': key.S,
    'T': key.T,
    'U': key.U,
    'V': key.V,
    'W': key.W,
    'X': key.X,
    'Y': key.Y,
    'Z': key.Z,
    '{': key.BRACELEFT,
    '|': key.BAR,
    '}': key.BRACERIGHT,
    '~': key.ASCIITILDE
}



"""Joystick, Game Controller, Tablet and USB HID device support.

This module provides a unified interface to almost any input device, besides
the regular mouse and keyboard support provided by
:py:class:`~pyglet.window.Window`.  At the lowest
level, :py:func:`get_devices` can be used to retrieve a list of all supported
devices, including joysticks, tablets, game controllers, wheels, pedals, remote
controls, keyboards and mice.  The set of returned devices varies greatly
depending on the operating system (and, of course, what's plugged in).  

At this level pyglet does not try to interpret *what* a particular device is,
merely what controls it provides.  A :py:class:`Control` can be either a button,
whose value is either ``True`` or ``False``, or a relative or absolute-valued
axis, whose value is a float.  Sometimes the name of a control can be provided
(for example, ``x``, representing the horizontal axis of a joystick), but often
not.  In these cases the device API may still be useful -- the user will have
to be asked to press each button in turn or move each axis separately to
identify them.

Higher-level interfaces are provided for joysticks, game controllers, tablets
and the Apple remote control. These devices can usually be identified by pyglet
positively, and a base level of functionality for each one provided through a
common interface.

To use an input device:

1. Call :py:func:`get_devices`, :py:func:`get_apple_remote`,
   :py:func:`get_controllers` or :py:func:`get_joysticks` to retrieve and
   identify the device.
2. For low-level devices (retrieved by :py:func:`get_devices`), query the
   devices list of controls and determine which ones you are interested in. For
   high-level interfaces the set of controls is provided by the interface.
3. Optionally attach event handlers to controls on the device. For high-level
   interfaces, additional events are available.
4. Call :py:meth:`Device.open` to begin receiving events on the device.  You can
   begin querying the control values after this time; they will be updated
   asynchronously.
5. Call :py:meth:`Device.close` when you are finished with the device (not
   needed if your application quits at this time).

To use a tablet, follow the procedure above using :py:func:`get_tablets`, but
note that no control list is available; instead, calling :py:meth:`Tablet.open`
returns a :py:class:`TabletCanvas` onto which you should set your event
handlers.

For game controllers, the :py:class:`ControllerManager` is available. This
provides a convenient way to handle hot-plugging of controllers.

.. versionadded:: 1.2

"""
from __future__ import annotations

import sys

from typing import TYPE_CHECKING

from .base import Device, Control, RelativeAxis, AbsoluteAxis, ControllerManager
from .base import Button, Joystick, AppleRemote, Tablet, Controller
from .base import DeviceException, DeviceOpenException, DeviceExclusiveException

if TYPE_CHECKING:
    from pyglet.display import Display

_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run


def get_apple_remote(display: Display | None = None) -> AppleRemote | None:
    """Get the Apple remote control device, if it exists.

    The Apple remote is the small white 6-button remote control that
    accompanies most recent Apple desktops and laptops.  The remote can only
    be used with Mac OS X.

    Args:
        display:
            Currently ignored.
    """
    return None


if _is_pyglet_doc_run:
    def get_devices(display: Display | None = None) -> list[Device]:
        """Get a list of all attached input devices.

        Args:
            display:
                The display device to query for input devices.  Ignored on Mac
                OS X and Windows.  On Linux, defaults to the default display device.
        """


    def get_joysticks(display: Display | None = None) -> list[Joystick]:
        """Get a list of attached joysticks.

        Args:
            display:
                The display device to query for input devices.  Ignored on Mac
                OS X and Windows.  On Linux, defaults to the default display device.
        """


    def get_controllers(display: Display | None = None) -> list[Controller]:
        """Get a list of attached controllers.

        Args:
            display:
                The display device to query for input devices.  Ignored on Mac
                OS X and Windows.  On Linux, defaults to the default display device.
        """

    def get_tablets(display: Display | None = None) -> list[Tablet]:
        """Get a list of tablets.

        This function may return a valid tablet device even if one is not
        attached (for example, it is not possible on Mac OS X to determine if
        a tablet device is connected).  Despite returning a list of tablets,
        pyglet does not currently support multiple tablets, and the behaviour
        is undefined if more than one is attached.

        Args:
            display:
                The display device to query for input devices.  Ignored on Mac
                OS X and Windows.  On Linux, defaults to the default display device.
        """

else:

    from pyglet import compat_platform

    if compat_platform.startswith('linux'):
        from .linux import get_devices
        from .linux import get_joysticks
        from .linux import get_controllers
        from .linux import get_tablets
        from .linux import ControllerManager

    elif compat_platform in ('cygwin', 'win32'):
        from .win32 import get_devices
        from .win32 import get_joysticks
        from .win32 import get_controllers
        from .win32 import get_tablets
        from .win32 import Win32ControllerManager as ControllerManager

    elif compat_platform == 'darwin':
        from .macos import get_devices
        from .macos import get_joysticks
        from .macos import get_apple_remote
        from .macos import get_controllers
        from .macos import get_tablets
        from .macos import ControllerManager

__all__ = [
    'get_devices',
    'get_joysticks',
    'get_controllers',
    'get_tablets',
    'get_apple_remote',
    'ControllerManager',
    'Device',
    'Control',
    'RelativeAxis',
    'AbsoluteAxis',
    'Button',
    'Joystick',
    'AppleRemote',
    'Tablet',
    'Controller',
    'DeviceException',
    'DeviceOpenException',
    'DeviceExclusiveException',
]



import os
import unittest

import pytest


local_dir = os.path.dirname(__file__)
test_data_path = os.path.abspath(os.path.join(local_dir, '..', 'data'))
del local_dir


class PygletTestCase(unittest.TestCase):
    """
    Base class for pyglet tests.
    Specifies helper methods for all tests.
    """
    @staticmethod
    def get_test_data_file(*file_parts):
        """
        Get a file from the test data directory in an OS independent way. Supply relative file
        name as you would in os.path.join().
        """
        return os.path.join(test_data_path, *file_parts)


class TestDataFixture:
    """Fixture for accessing test data."""
    def __init__(self):
        local_dir = os.path.dirname(__file__)
        self._test_data_path = os.path.abspath(os.path.join(local_dir, '..', 'data'))

    def get_file(self, *file_parts):
        """
        Get a file from the test data directory in an OS independent way. Supply relative file
        name as you would in os.path.join().
        """
        return os.path.join(self._test_data_path, *file_parts)


@pytest.fixture(scope="session")
def test_data():
    return TestDataFixture()



"""Mouse constants and utilities for pyglet.window."""
from __future__ import annotations


class MouseStateHandler:
    """Simple handler that tracks the state of buttons and coordinates from the mouse.

    If a button is pressed then this handler holds a ``True`` value for it.
    If the window loses focus, all values will be reset to ``False`` in order
    to avoid a "sticky" state.

    For example::

        >>> win = window.Window()
        >>> mouse_state = mouse.MouseStateHandler()
        >>> win.push_handlers(mouse_state)

        # Hold down the "left" button...

        >>> mouse_state[mouse.LEFT]
        True
        >>> mouse_state[mouse.RIGHT]
        False


    Mouse coordinates can be retrieved by using the ``'x'`` and ``'y'`` strings
    or by using ``'mouse_state.x'`` and ``'mouse_state.y`` attributes.

    For example::

        >>> win = window.Window()
        >>> mouse_state = mouse.MouseStateHandler()
        >>> win.push_handlers(mouse_state)

        # Move the mouse around...

        >>> mouse_state['x']
        20
        >>> mouse_state['y']
        50
        
        # Or...
        >>> mouse_state.x
        20
        >>> mouse_state.y
        50
    """

    def __init__(self) -> None:  # noqa: D107
        self.data = {
            'x': 0,
            'y': 0,
        }

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> None:  # noqa: ARG002
        self.data[button] = True

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int) -> None:  # noqa: ARG002
        self.data[button] = False

    def on_deactivate(self) -> None:
        self.data.clear()

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:  # noqa: ARG002
        self.data['x'] = x
        self.data['y'] = y

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:  # noqa: ARG002
        self.data['x'] = x
        self.data['y'] = y

    def __getitem__(self, key: str) -> int | bool:
        return self.data.get(key, False)
    
    def __getattr__(self, item: str) -> int:
        return self.data[item]


def buttons_string(buttons: int) -> str:
    """Return a string describing a set of active mouse buttons.

    Example::

        >>> buttons_string(LEFT | RIGHT)
        'LEFT|RIGHT'

    Args:
        buttons:
            Bitwise combination of mouse button constants.
    """
    button_names = []
    if buttons & LEFT:
        button_names.append('LEFT')
    if buttons & MIDDLE:
        button_names.append('MIDDLE')
    if buttons & RIGHT:
        button_names.append('RIGHT')
    if buttons & MOUSE4:
        button_names.append('MOUSE4')
    if buttons & MOUSE5:
        button_names.append('MOUSE5')
    return '|'.join(button_names)


#: Constant for the left mouse button.
#:
#: :meta hide-value:
LEFT = 1 << 0
#: Constant for the middle mouse button.
#:
#: :meta hide-value:
MIDDLE = 1 << 1
#: Constant for the right mouse button.
#:
#: :meta hide-value:
RIGHT = 1 << 2
#: Constant for the mouse4 button.
#:
#: :meta hide-value:
MOUSE4 = 1 << 3
#: Constant for the mouse5 button.
#:
#: :meta hide-value:
MOUSE5 = 1 << 4



#!/usr/bin/env python
"""A package for generating ctypes wrappers from a C header file.

See the docstring for wrap.py for usage.

The files 'lex.py' and 'yacc.py' are from PLY (http://www.dabeaz.com/ply),
which was written by David M. Beazley but have been modified slightly
for this tool.
"""


import pytest

from pyglet import resource


@pytest.mark.parametrize('transforms,tex_order', [
    (dict(), (0, 1, 2, 3)),
    (dict(flip_x=True), (1, 0, 3, 2)),
    (dict(flip_y=True), (3, 2, 1, 0)),
    (dict(flip_x=True, flip_y=True), (2, 3, 0, 1)),
    (dict(rotate=90), (1, 2, 3, 0)),
    (dict(rotate=-270), (1, 2, 3, 0)),
    (dict(rotate=180), (2, 3, 0, 1)),
    (dict(rotate=-180), (2, 3, 0, 1)),
    (dict(rotate=270), (3, 0, 1, 2)),
    (dict(rotate=-90), (3, 0, 1, 2)),
])
def test_resource_image_loading(event_loop, transforms, tex_order):
    """Test loading an image resource with possible transformations."""
    resource.path.append('@' + __name__)
    resource.reindex()

    tex = resource.image('rgbm.png', **transforms)

    assert tex.tex_coords_order == tex_order, f"{transforms}, {tex.tex_coords_order} != {tex_order}"



import itertools
import inspect
from typing import Dict, Optional, Generator, Tuple, Type

import pyglet.window
import math
from pyglet.shapes import (
    Arc,
    BezierCurve,
    Circle,
    Ellipse,
    Sector,
    Line,
    MultiLine,
    Rectangle,
    BorderedRectangle,
    Triangle,
    Star,
    Polygon,
    Box,
    ShapeBase
)
from pyglet.text import Label

INSTRUCTIONS = """- / = keys to rotate
Arrow keys to change anchor points
Left click + drag to adjust positions
"""

PX_SIZE = 10
PX_SIZE_2 = PX_SIZE * 2

# Call next() to get a new color; endlessly repeats
colors = itertools.cycle(map(lambda c: c + (255,), [
    (255, 0, 0),  # Red
    (0, 255, 0),  # Blue
    (0, 0, 255),  # Green
    (255, 255, 0)  # Yellow
]))

# Default values for each shape type: Type, *args, **kwargs
shape_configs = [
    (Arc, (PX_SIZE,), {}),
    (BezierCurve, [(0, 0), (PX_SIZE, PX_SIZE_2), (PX_SIZE_2, PX_SIZE)], {}),
    (Circle, (PX_SIZE,), {}),
    (Ellipse, (PX_SIZE, PX_SIZE_2,), {}),

    (Sector, (PX_SIZE,), {}),
    (Line, (0, 0, PX_SIZE_2, PX_SIZE_2), {}),
    (Rectangle, (PX_SIZE, PX_SIZE_2), {}),
    (BorderedRectangle, (PX_SIZE_2, PX_SIZE), {}),
    (Box, (PX_SIZE_2, PX_SIZE_2), {}),

    (Triangle, (-PX_SIZE, 0, 0, PX_SIZE, PX_SIZE, 0), {}),
    (Star, (PX_SIZE_2, PX_SIZE, 5), {}),
    (Polygon, ((-PX_SIZE, 0), (0, PX_SIZE), (PX_SIZE, 0)), {}),
    (MultiLine, ((-PX_SIZE, 0), (0, PX_SIZE), (PX_SIZE, 0)), {}),
]


def layout_points(
        width: int, height: int,
        padding: int,
        num_cols: int,
        num_rows: Optional[int] = None
) -> Generator[Tuple[int, int], None, None]:
    """Yield points out in a col / row grid.

    All units in px.
    Args:
        width: Width of the area
        height: Height of the area
        padding: How many px to pad on all sides
        num_cols: How many columns each row should have
        num_rows: Specify to prevent auto-calculation

    Yields:
        Bottom left screen position of the grid cell in px
    """
    x_step_range = width - 2 * padding
    y_step_range = height - 2 * padding

    if num_rows is None:
        num_rows = math.ceil(len(shape_configs) / num_cols)

    x_step = x_step_range // num_cols
    y_step = y_step_range // num_rows

    for y in range(height - padding, y_step, -y_step):
        for x in range(padding, padding + x_step_range, x_step):
            yield x, y


def main():
    padding = 50
    width, height = 600, 600

    window = pyglet.window.Window(
        width, height,
        caption="Shape Translation / Rotation Spotcheck",
        resizable=True)

    shapes: Dict[Type, ShapeBase] = {}
    labels: Dict[Type, Label] = {}
    batch = pyglet.graphics.Batch()

    for shape_config, position in zip(shape_configs, layout_points(window.width, window.height, padding, 4)):
        shape_type, args, kwargs = shape_config
        sig = inspect.signature(shape_type)
        params = list(sig.parameters)
        print(shape_type.__name__, params)

        # Arrange shape args
        if params[:2] == ['x', 'y'] and params[2] != 'x2':
            final_args = position + args
        elif shape_type in (Polygon, BezierCurve, MultiLine):
            raw = [(position[0] + p[0], position[1] + p[1]) for p in args]
            final_args = raw
            print(shape_type.__name__, final_args)

        elif shape_type in (Line, Triangle):
            final_args = [
                position[0] + args[0], position[1] + args[1],
                position[0] + args[2], position[1] + args[3]]
            if shape_type is Triangle:
                final_args += [position[0] + args[4], position[1] + args[5]]
            print(shape_type.__name__, final_args)

        else:
            raise TypeError("unhandled")

        final_kwargs = dict(color=next(colors))
        final_kwargs.update(kwargs)

        label = pyglet.text.Label(
            shape_type.__name__,
            x=position[0], y=position[1] - padding,
            batch=batch,
            anchor_x='center',
            **final_kwargs
        )
        shape = shape_type(*final_args, batch=batch, **final_kwargs)
        shapes[shape_type] = shape
        labels[shape_type] = label

    _instructions = pyglet.text.Label(
        INSTRUCTIONS,
        x=window.width // 2, y=padding,
        width=window.width,
        anchor_x='center',
        align='center',
        batch=batch,
        multiline=True
    )

    @window.event
    def on_draw():
        window.clear()
        batch.draw()

    @window.event
    def on_key_press(symbol, modifiers):
        if symbol == pyglet.window.key.MINUS:
            for item in shapes.values():
                item.rotation -= 20
        elif symbol == pyglet.window.key.EQUAL:
            for item in shapes.values():
                item.rotation += 20
        elif symbol == pyglet.window.key.LEFT:
            for item in shapes.values():
                item.anchor_x += 10
        elif symbol == pyglet.window.key.RIGHT:
            for item in shapes.values():
                item.anchor_x -= 10
        elif symbol == pyglet.window.key.DOWN:
            for item in shapes.values():
                item.anchor_y += 10
        elif symbol == pyglet.window.key.UP:
            for item in shapes.values():
                item.anchor_y -= 10

    @window.event
    def on_mouse_drag(x, y, dx, dy, buttons, modifiers):
        if buttons == pyglet.window.mouse.LEFT:
            for item in itertools.chain(shapes.values(), labels.values()):
                item_x, item_y, *item_z = item.position
                item.position = item_x + dx, item_y + dy, *item_z

    # This line is crucial for 2.1 / the development branch
    pyglet.clock.schedule_interval(window.draw, 1 / 60)
    pyglet.app.run()


if __name__ == "__main__":
    main()



import pytest
import time

from tests.annotations import Platform, require_platform
from tests.base.interactive import InteractiveTestCase
from tests.interactive.window import window_util

from pyglet import window
from pyglet.gl import *
from pyglet.window.event import WindowEventLogger

@pytest.mark.requires_user_validation
class WINDOW_MINIMIZE_MAXIMIZE(InteractiveTestCase):
    """Test that window can be minimized and maximized.

    Expected behaviour:
        One window will be opened. It will be maximized and minimized.
    """
    def test_minimize_maximize(self):
        self.width, self.height = 200, 200
        self.w = w = window.Window(self.width, self.height, resizable=True)
        try:
            w.dispatch_events()
            self.user_verify('Is the window visible and not maximized?',
                             take_screenshot=False)

            w.maximize()
            w.dispatch_events()
            self.user_verify('Is the window maximized?',
                             take_screenshot=False)

            w.minimize()
            w.dispatch_events()
            self.user_verify('Is the window minimized?',
                             take_screenshot=False)

        finally:
            w.close()


@pytest.mark.requires_user_action
class WINDOW_ACTIVATE(InteractiveTestCase):
    """Test that the window can be activated (focus set).

    Expected behaviour:
        One window will be opened.  Every 5 seconds it will be activated;
        it should be come to the front and accept keyboard input (this will
        be shown on the terminal).

        On some OSes, the taskbar icon may flash (indicating the application
        requires attention) rather than moving the window to the foreground.  This
        is the correct behaviour.

        Press escape or close the window to finished the test.
    """
    def test_activate(self):
        w = window.Window(200, 200)
        try:
            w.push_handlers(WindowEventLogger())
            last_time = time.time()
            while not w.has_exit:
                if time.time() - last_time > 5:
                    w.activate()
                    last_time = time.time()
                    print('Activated window.')
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_RESIZABLE(InteractiveTestCase):
    """Test that window can be resized.

    Expected behaviour:
        One window will be opened.  It should be resizable by the user. 

        Close the window or press ESC to end the test.
    """
    def test_resizable(self):
        self.width, self.height = 200, 200
        self.w = w = window.Window(self.width, self.height, resizable=True)
        try:
            glClearColor(1, 1, 1, 1)
            while not w.has_exit:
                window_util.draw_client_border(w)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
@require_platform(Platform.WINDOWS + Platform.OSX)
class WINDOW_MODE_SWITCH(InteractiveTestCase):
    """Test switching to available screen modes."""
    def on_text(self, text):
        text = text[:1]
        i = ord(text) - ord('a')
        if 0 <= i < len(self.modes):
            print('Switching to %s' % self.modes[i])
            self.w.screen.set_mode(self.modes[i])

    def on_expose(self):
        glClearColor(1, 0, 0, 1)
        glClear(GL_COLOR_BUFFER_BIT)
        window_util.draw_client_border(self.w)
        self.w.flip()

    def test_set_fullscreen(self):
        self.w = w = window.Window(200, 200)
        try:
            self.modes = w.screen.get_modes()
            self.assertTrue(len(self.modes) > 0, msg='No modes available')
            print('Press a letter to switch to the corresponding mode:')
            for i, mode in enumerate(self.modes):
                print('%s: %s' % (chr(i + ord('a')), mode))

            w.push_handlers(self)
            self.on_expose()
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)



import pyglet
import warnings

from .base import Display, Screen, ScreenMode, Canvas


from ctypes import byref
from pyglet.libs.egl import egl
from pyglet.libs.egl import eglext


class HeadlessDisplay(Display):

    def __init__(self):
        super().__init__()
        # TODO: fix this placeholder:
        self._screens = [HeadlessScreen(self, 0, 0, 1920, 1080)]

        num_devices = egl.EGLint()
        eglext.eglQueryDevicesEXT(0, None, byref(num_devices))
        if num_devices.value > 0:
            headless_device = pyglet.options['headless_device']
            if headless_device < 0 or headless_device >= num_devices.value:
                raise ValueError(f'Invalid EGL device id: {headless_device}')
            devices = (eglext.EGLDeviceEXT * num_devices.value)()
            eglext.eglQueryDevicesEXT(num_devices.value, devices, byref(num_devices))
            self._display_connection = eglext.eglGetPlatformDisplayEXT(
                eglext.EGL_PLATFORM_DEVICE_EXT, devices[headless_device], None)
        else:
            warnings.warn('No device available for EGL device platform. Using native display type.')
            display = egl.EGLNativeDisplayType()
            self._display_connection = egl.eglGetDisplay(display)

        egl.eglInitialize(self._display_connection, None, None)

    def get_screens(self):
        return self._screens

    def __del__(self):
        egl.eglTerminate(self._display_connection)


class HeadlessCanvas(Canvas):
    def __init__(self, display, egl_surface):
        super().__init__(display)
        self.egl_surface = egl_surface


class HeadlessScreen(Screen):
    def __init__(self, display, x, y, width, height):
        super().__init__(display, x, y, width, height)

    def get_matching_configs(self, template):
        canvas = HeadlessCanvas(self.display, None)
        configs = template.match(canvas)
        # XXX deprecate
        for config in configs:
            config.screen = self
        return configs

    def get_modes(self):
        pass

    def get_mode(self):
        pass

    def set_mode(self, mode):
        pass

    def restore_mode(self):
        pass



import random

import pyglet

window = pyglet.window.Window(vsync=False)

# Set example resource path.
pyglet.resource.path = ['../resources']
pyglet.resource.reindex()

# Load example image from resource path.
# Avoid using pyglet.image.load as it creates multiple textures.
# Resource will load images into a texture atlas, allowing significantly faster performance.
image = pyglet.resource.image("pyglet.png")

# Anchor point on an image is bottom left corner by default.
# Set to center point with anchor properties.
image.anchor_x = image.width // 2
image.anchor_y = image.height // 2

# Batching allows rendering groups of objects all at once instead of drawing one by one.
batch = pyglet.graphics.Batch()

scales = [1.0, 0.75, 0.5, 0.25]

sprites = []
# Create 1000 sprites at various scales.
for i in range(1000):
    sprite = pyglet.sprite.Sprite(image,
                                  x=random.randint(0, window.width),
                                  y=random.randint(0, window.height),
                                  batch=batch)  # specify the batch to enter the sprites in.

    # Randomize scale.
    sprite.scale = random.choice(scales)

    # Random color multiplier.
    sprite.color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

    # Add sprites to keep in memory, like a list. Otherwise they will get GC'd when out of scope.
    sprites.append(sprite)


@window.event
def on_draw():
    window.clear()
    batch.draw()


pyglet.app.run()



"""Tests for loading and accessing FreeType font faces."""
import pytest

from tests.annotations import Platform, require_platform

try:
    from pyglet.font.freetype import FreeTypeFace, FreeTypeMemoryFace
    from pyglet.font.fontconfig import get_fontconfig
except ImportError:
    FreeTypeFace = None
    FreeTypeMemoryFace = None
    get_fontconfig = None


@require_platform(Platform.LINUX)
@pytest.mark.parametrize('font_file_name,font_name,bold,italic', [
    ('action_man.ttf', 'Action Man', False, False),
    ('action_man_bold.ttf', 'Action Man', True, False),
    ('action_man_bold_italic.ttf', 'Action Man', True, True),
    ('action_man_italic.ttf', 'Action Man', False, True),
    ])
def test_face_from_file(test_data, font_file_name, font_name, bold, italic):
    """Test loading a font face directly from file."""
    face = FreeTypeFace.from_file(test_data.get_file('fonts', font_file_name))

    assert face.name == font_name
    assert face.family_name == font_name
    assert face.bold == bold
    assert face.italic == italic

    del face


@require_platform(Platform.LINUX)
@pytest.mark.parametrize('font_name,bold,italic', [
    ('Arial', False, False),
    ('Arial', True, False),
    ('Arial', False, True),
    ('Arial', True, True),
    ])
def test_face_from_fontconfig(font_name, bold, italic):
    """Test loading a font face from the system using font config."""
    match = get_fontconfig().find_font(font_name, 16, bold, italic)
    assert match is not None

    face = FreeTypeFace.from_fontconfig(match)

    assert face.bold == bold
    assert face.italic == italic

    del face


@require_platform(Platform.LINUX)
@pytest.mark.parametrize('font_file_name,font_name,bold,italic', [
    ('action_man.ttf', 'Action Man', False, False),
    ('action_man_bold.ttf', 'Action Man', True, False),
    ('action_man_bold_italic.ttf', 'Action Man', True, True),
    ('action_man_italic.ttf', 'Action Man', False, True),
    ])
def test_memory_face(test_data, font_file_name, font_name, bold, italic):
    """Test loading a font into memory using FreeTypeMemoryFont."""
    with open(test_data.get_file('fonts', font_file_name), 'rb') as font_file:
        font_data = font_file.read()
    font = FreeTypeMemoryFace(font_data, 0)

    assert font.name == font_name
    assert font.bold == bold
    assert font.italic == italic
    assert font.ft_face is not None

    del font


@require_platform(Platform.LINUX)
@pytest.mark.parametrize('font_file_name,size,dpi,ascent,descent', [
    ('action_man.ttf', 16, 96, 15, -4),
    ('action_man.ttf', 10, 96, 9, -3),
    ('action_man.ttf', 16, 72, 11, -3),
    ('courR12-ISO8859-1.pcf', 16, 96, 12, -3),

    # A fixed bitmap size font. Will fail if render size is used. Keep for legacy info.
    # ('courR12-ISO8859-1.pcf', 16, 96, 15, -4),
    ])
def test_face_metrics(test_data, font_file_name, size, dpi, ascent, descent):
    """Test a face from file and check the metrics."""
    face = FreeTypeFace.from_file(test_data.get_file('fonts', font_file_name))
    metrics = face.get_font_metrics(size, dpi)

    assert metrics.ascent == ascent
    assert metrics.descent == descent


@require_platform(Platform.LINUX)
@pytest.mark.parametrize('font_file_name,character,index', [
    ('action_man.ttf', 'a', 65),
    ('action_man.ttf', 'A', 33),
    ('action_man.ttf', '1', 17),
    ('action_man.ttf', '#', 3),
    ('action_man.ttf', 'b', 66),
    ])
def test_character_index(test_data, font_file_name, character, index):
    """Test getting the glyph index for a character."""
    face = FreeTypeFace.from_file(test_data.get_file('fonts', font_file_name))
    assert face.get_character_index(character) == index


@require_platform(Platform.LINUX)
@pytest.mark.parametrize('font_file_name,glyph_index', [
    ('action_man.ttf', 65),
    ('action_man.ttf', 33),
    ('action_man.ttf', 17),
    ('action_man.ttf', 3),
    ('action_man.ttf', 66),
    ])
def test_get_glyph_slot(test_data, font_file_name, glyph_index):
    """Test getting a glyph slot from the face."""
    face = FreeTypeFace.from_file(test_data.get_file('fonts', font_file_name))
    face.set_char_size(size=16, dpi=92)
    glyph = face.get_glyph_slot(glyph_index)

    assert glyph is not None




"""Minimal Compute Shader example.

This example creates a Compute Shader from source,
attached a texture to it, and writes into the texture.
The texture is then saved to disk as a .png file.
"""

import pyglet

from pyglet.gl import GL_RGBA32F, GL_ALL_BARRIER_BITS


compute_src = """#version 430 core
layout (local_size_x = 1, local_size_y = 1, local_size_z = 1) in;

layout(rgba32f) uniform image2D img_output;

void main() {
    vec4 value = vec4(0.0, 0.0, 0.0, 1.0);
    ivec2 texel_coord = ivec2(gl_GlobalInvocationID.xy);
    value.r = float(texel_coord.x)/(gl_NumWorkGroups.x);
    value.g = float(texel_coord.y)/(gl_NumWorkGroups.y);

    imageStore(img_output, texel_coord, value);
}
"""

program = pyglet.graphics.shader.ComputeShaderProgram(compute_src)

# Create an RGBA32F Texture that we can bind to the ShaderProgram.
out_texture = pyglet.image.Texture.create(540, 540, internalformat=GL_RGBA32F)

# The image binding unit in the shader needs to be set for the texture.
# The uniform set for the image can be used. In most cases this will be 0 unless explicitly
# changed via layout qualifier of layout(binding=X, rgba32f)
output_binding = program['img_output']
out_texture.bind_image_texture(unit=output_binding)


with program:
    program.dispatch(out_texture.width, out_texture.height, 1, barrier=GL_ALL_BARRIER_BITS)


out_texture.save('compute_output.png')
print("Saved 'compute_output.png'")



from __future__ import annotations

import pyglet

from pyglet.input import base
from pyglet.input.win32.directinput import _di_manager as _di_device_manager
from pyglet.input.win32.directinput import DirectInputDevice, _create_controller

from pyglet.input.win32.directinput import get_devices as dinput_get_devices
from pyglet.input.win32.directinput import get_controllers as dinput_get_controllers
from pyglet.input.win32.directinput import get_joysticks

try:
    from pyglet.input.win32.wintab import get_tablets
except:
    def get_tablets(display=None):
        import warnings
        warnings.warn("Failed to initialize wintab framework.")
        return []


_xinput_enabled = False
if not pyglet.options["win32_disable_xinput"]:
    try:
        from pyglet.input.win32.xinput import XInputControllerManager, XInputController, XInputDevice
        from pyglet.input.win32.xinput import _device_manager as _xinput_device_manager
        from pyglet.input.win32.xinput import get_devices as xinput_get_devices
        from pyglet.input.win32.xinput import get_controllers as xinput_get_controllers

        _xinput_enabled = True
    except OSError:
        # Fail to import XInput.
        pass


if not _xinput_enabled:
    def xinput_get_devices():
        return []

    def xinput_get_controllers():
        return []


class Win32ControllerManager(base.ControllerManager):
    """This class manages XInput and DirectInput as a combined manager.
       XInput will override any XInput compatible DirectInput devices.
       Any devices not supported by XInput will fall back to DirectInput.
    """

    def __init__(self):
        self._di_controllers: dict[DirectInputDevice, base.Controller] = {}

        if _xinput_enabled:
            self._xinput_controllers: dict[XInputDevice, XInputController] = {}

            for xdev in _xinput_device_manager.all_devices:  # All 4 devices are initialized.
                meta = {'name': xdev.name, 'guid': "XINPUTCONTROLLER"}
                self._xinput_controllers[xdev] = XInputController(xdev, meta)

            @_xinput_device_manager.event
            def on_connect(xdevice):
                self.dispatch_event('on_connect', self._xinput_controllers[xdevice])

            @_xinput_device_manager.event
            def on_disconnect(xdevice):
                self.dispatch_event('on_disconnect', self._xinput_controllers[xdevice])

        self._set_initial_didevices()

        @_di_device_manager.event
        def on_connect(di_device):
            if di_device not in self._di_controllers:
                if self._add_di_controller(di_device):
                    pyglet.app.platform_event_loop.post_event(self, 'on_connect', self._di_controllers[di_device])

        @_di_device_manager.event
        def on_disconnect(di_device):
            if di_device in self._di_controllers:
                _controller = self._di_controllers[di_device]
                del self._di_controllers[di_device]
                pyglet.app.platform_event_loop.post_event(self, 'on_disconnect', _controller)

    def _set_initial_didevices(self):
        if not _di_device_manager.registered:
            _di_device_manager.register_device_events()
            _di_device_manager.set_current_devices()

        for device in _di_device_manager.devices:
            self._add_di_controller(device)

    def _add_di_controller(self, device: DirectInputDevice) -> None | base.Controller:
        controller = _create_controller(device)
        if controller:
            self._di_controllers[device] = controller
            return controller

        return None

    def _get_xinput_controllers(self) -> list:
        if not _xinput_enabled:
            return []
        return [ctlr for ctlr in self._xinput_controllers.values() if ctlr.device.connected]

    def _get_di_controllers(self) -> list:
        return list(self._di_controllers.values())

    def get_controllers(self):
        return self._get_xinput_controllers() + self._get_di_controllers()


def get_devices(display=None):
    return xinput_get_devices() + dinput_get_devices(display)


def get_controllers(display=None):
    return xinput_get_controllers() + dinput_get_controllers(display)



# Separate some structure types from the lib to prevent circular imports, as dwrite/d2d render targets use these too.
from __future__ import annotations

from ctypes import Structure
from ctypes.wintypes import FLOAT, LONG

from pyglet.libs.win32 import UINT32


class D2D1_SIZE_F(Structure):
    _fields_ = [
        ("width", FLOAT),
        ("height", FLOAT),
    ]


class D2D1_SIZE_U(Structure):
    _fields_ = [
        ("width", UINT32),
        ("height", UINT32),
    ]

class D2D1_POINT_2L(Structure):
    _fields_ = [
        ("x", LONG),
        ("y", LONG),
    ]

class D2D1_POINT_2F(Structure):
    _fields_ = [
        ("x", FLOAT),
        ("y", FLOAT),
    ]

class D2D1_COLOR_F(Structure):
    _fields_ = (
        ("r", FLOAT),
        ("g", FLOAT),
        ("b", FLOAT),
        ("a", FLOAT),
    )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(r={self.r}, g={self.g}, b={self.b}, a={self.a})"


class D2D_POINT_2F(Structure):
    _fields_ = (
        ("x", FLOAT),
        ("y", FLOAT),
    )


class D2D1_RECT_F(Structure):
    _fields_ = (
        ("left", FLOAT),
        ("top", FLOAT),
        ("right", FLOAT),
        ("bottom", FLOAT),
    )



"""Get environment information useful for debugging.

Intended usage is to create a file for bug reports, e.g.::

    python -m pyglet.info > info.txt

"""

_first_heading = True


def _heading(heading):
    global _first_heading
    if not _first_heading:
        print()
    else:
        _first_heading = False
    print(heading)
    print('-' * 78)


def dump_platform():
    """Dump OS specific """
    import platform
    print('platform: ', platform.platform())
    print('release:  ', platform.release())
    print('machine:  ', platform.machine())


def dump_python():
    """Dump Python version and environment to stdout."""
    import os
    import sys
    import platform
    print('implementation:', platform.python_implementation())
    print('sys.version:', sys.version)
    print('sys.maxint:', sys.maxsize)
    if sys.platform == 'darwin':
        try:
            from objc import __version__ as pyobjc_version
            print('objc.__version__:', pyobjc_version)
        except:
            print('PyObjC not available')
    print('os.getcwd():', os.getcwd())
    for key, value in os.environ.items():
        if key.startswith('PYGLET_'):
            print(f"os.environ['{key}']: {value}")


def dump_pyglet():
    """Dump pyglet version and options."""
    import pyglet
    print('pyglet.version:', pyglet.version)
    print('pyglet.compat_platform:', pyglet.compat_platform)
    print('pyglet.__file__:', pyglet.__file__)
    for key, value in pyglet.options.items():
        print(f"pyglet.options['{key}'] = {value!r}")


def dump_window():
    """Dump display, window, screen and default config info."""
    from pyglet.gl import gl_info
    if not gl_info.have_version(3):
        print(f"Insufficient OpenGL version: {gl_info.get_version_string()}")
        return
    import pyglet.window
    display = pyglet.display.get_display()
    print('display:', repr(display))
    screens = display.get_screens()
    for i, screen in enumerate(screens):
        print(f'screens[{i}]: {screen!r}')
    window = pyglet.window.Window(visible=False)
    for key, value in window.config.get_gl_attributes():
        print(f"config['{key}'] = {value!r}")
    print('context:', repr(window.context))

    _heading('window.context._info')
    dump_gl(window.context)
    window.close()


def dump_gl(context=None):
    """Dump GL info."""
    if context is not None:
        info = context.get_info()
    else:
        from pyglet.gl import gl_info as info
    print('gl_info.get_version():', info.get_version())
    print('gl_info.get_vendor():', info.get_vendor())
    print('gl_info.get_renderer():', info.get_renderer())


def dump_glx():
    """Dump GLX info."""
    try:
        from pyglet.gl import glx_info
    except:
        print('GLX not available.')
        return
    import pyglet
    window = pyglet.window.Window(visible=False)
    print('context.is_direct():', window.context.is_direct())
    window.close()

    if not glx_info.have_version(1, 1):
        print('Version: < 1.1')
    else:
        print('glx_info.get_server_vendor():', glx_info.get_server_vendor())
        print('glx_info.get_server_version():', glx_info.get_server_version())
        print('glx_info.get_server_extensions():')
        for name in glx_info.get_server_extensions():
            print('  ', name)
        print('glx_info.get_client_vendor():', glx_info.get_client_vendor())
        print('glx_info.get_client_version():', glx_info.get_client_version())
        print('glx_info.get_client_extensions():')
        for name in glx_info.get_client_extensions():
            print('  ', name)
        print('glx_info.get_extensions():')
        for name in glx_info.get_extensions():
            print('  ', name)


def dump_media():
    """Dump pyglet.media info."""
    import pyglet.media
    print('audio driver:', pyglet.media.get_audio_driver())


def dump_ffmpeg():
    """Dump FFmpeg info."""
    import pyglet
    pyglet.options['search_local_libs'] = True
    import pyglet.media

    if pyglet.media.have_ffmpeg():
        from pyglet.media.codecs.ffmpeg import get_version
        print('FFmpeg version:', get_version())
    else:
        print('FFmpeg not available.')


def dump_al():
    """Dump OpenAL info."""
    try:
        from pyglet.media.drivers import openal
    except:
        print('OpenAL not available.')
        return
    print('Library:', openal.lib_openal._lib)

    driver = openal.create_audio_driver()
    print('Version: {}.{}'.format(*driver.get_version()))
    print('Extensions:')
    for extension in driver.get_extensions():
        print('  ', extension)


def dump_wintab():
    """Dump WinTab info."""
    try:
        from pyglet.input.win32 import wintab
    except:
        print('WinTab not available.')
        return

    interface_name = wintab.get_interface_name()
    impl_version = wintab.get_implementation_version()
    spec_version = wintab.get_spec_version()

    print('WinTab: {0} {1}.{2} (Spec {3}.{4})'.format(interface_name,
                                                      impl_version >> 8, impl_version & 0xff,
                                                      spec_version >> 8, spec_version & 0xff))


def _try_dump(heading, func):
    _heading(heading)
    try:
        func()
    except:
        import traceback
        traceback.print_exc()


def dump():
    """Dump all information to stdout."""
    _try_dump('Platform', dump_platform)
    _try_dump('Python', dump_python)
    _try_dump('pyglet', dump_pyglet)
    _try_dump('pyglet.window', dump_window)
    _try_dump('pyglet.gl.glx_info', dump_glx)
    _try_dump('pyglet.media', dump_media)
    _try_dump('pyglet.media.ffmpeg', dump_ffmpeg)
    _try_dump('pyglet.media.drivers.openal', dump_al)
    _try_dump('pyglet.input.wintab', dump_wintab)


if __name__ == '__main__':
    dump()



import math


def distance(point_1=(0, 0), point_2=(0, 0)):
    """Returns the distance between two points"""
    return math.sqrt((point_1[0] - point_2[0]) ** 2 + (point_1[1] - point_2[1]) ** 2)


def center_image(image):
    """Sets an image's anchor point to its center"""
    image.anchor_x = image.width / 2
    image.anchor_y = image.height / 2



"""Print all Apple Remote events to stdout.
"""

import pyglet
import sys

window = pyglet.window.Window()


@window.event
def on_draw():
    window.clear()


remote = pyglet.input.get_apple_remote()
if not remote:
    print('Apple IR Remote not available.')
    sys.exit(0)

remote.open(window, exclusive=True)


@remote.select_control.event
def on_press():
    print('Press select')


@remote.menu_control.event
def on_press():
    print('Press menu')


@remote.up_control.event
def on_press():
    print('Press up')


@remote.down_control.event
def on_press():
    print('Press down')


@remote.left_control.event
def on_press():
    print('Press left')


@remote.right_control.event
def on_press():
    print('Press right')


@remote.select_control.event
def on_release():
    print('Release select')


@remote.menu_control.event
def on_release():
    print('Release menu')


@remote.up_control.event
def on_release():
    print('Release up')


@remote.down_control.event
def on_release():
    print('Release down')


@remote.left_control.event
def on_release():
    print('Release left')


@remote.right_control.event
def on_release():
    print('Release right')


pyglet.app.run()



from __future__ import annotations

from ctypes import c_bool, c_void_p
from typing import TYPE_CHECKING

from pyglet.libs.darwin.cocoapy import (
    NSRectEncoding,
    NSUInteger,
    NSUIntegerEncoding,
    ObjCClass,
    ObjCSubclass,
    send_super,
)

if TYPE_CHECKING:
    from pyglet.libs.darwin import ObjCInstance


class PygletWindow_Implementation:
    PygletWindow = ObjCSubclass('NSWindow', 'PygletWindow')

    @PygletWindow.method('B')
    def canBecomeKeyWindow(self) -> bool:
        return True

    # When the window is being resized, it enters into a mini event loop that
    # only looks at mouseDragged and mouseUp events, blocking everything else.
    # Among other things, this makes it impossible to run an NSTimer to call the
    # idle() function in order to update the view during the resize.  So we
    # override this method, called by the resizing event loop, and call the
    # idle() function from here.  This *almost* works.  I can't figure out what
    # is happening at the very beginning of a resize event.  The NSView's
    # viewWillStartLiveResize method is called and then nothing happens until
    # the mouse is dragged.  I think NSApplication's nextEventMatchingMask_etc
    # method is being called instead of this one.  I don't really feel like
    # subclassing NSApplication just to fix this.  Also, to prevent white flashes
    # while resizing, we must also call idle() from the view's reshape method.
    @PygletWindow.method(b'@' + NSUIntegerEncoding + b'@@B')
    def nextEventMatchingMask_untilDate_inMode_dequeue_(self, mask: ObjCInstance, date: ObjCInstance,
                                                        mode: ObjCInstance, dequeue: bool) -> int:
        if self.inLiveResize():
            # Call the idle() method while we're stuck in a live resize event.
            from pyglet import app
            if app.event_loop is not None:
                app.event_loop.idle()

        event = send_super(self, 'nextEventMatchingMask:untilDate:inMode:dequeue:',
                           mask, date, mode, dequeue,
                           superclass_name='NSWindow',
                           argtypes=[NSUInteger, c_void_p, c_void_p, c_bool])

        if event.value is None:
            return 0

        return event.value

    # Need this for set_size to not flash.
    @PygletWindow.method(b'd' + NSRectEncoding)
    def animationResizeTime_(self, newFrame: ObjCInstance) -> float:
        return 0.0


class PygletToolWindow_Implementation:
    PygletToolWindow = ObjCSubclass('NSPanel', 'PygletToolWindow')

    @PygletToolWindow.method(b'@' + NSUIntegerEncoding + b'@@B')
    def nextEventMatchingMask_untilDate_inMode_dequeue_(self, mask: ObjCInstance, date: ObjCInstance,
                                                        mode: ObjCInstance, dequeue: bool) -> int:
        if self.inLiveResize():
            # Call the idle() method while we're stuck in a live resize event.
            from pyglet import app
            if app.event_loop is not None:
                app.event_loop.idle()

        event = send_super(self, 'nextEventMatchingMask:untilDate:inMode:dequeue:',
                           mask, date, mode, dequeue, argtypes=[NSUInteger, c_void_p, c_void_p, c_bool])

        if event.value is None:
            return 0

        return event.value

    # Need this for set_size to not flash.
    @PygletToolWindow.method(b'd' + NSRectEncoding)
    def animationResizeTime_(self, newFrame: ObjCInstance) -> float:
        return 0.0


PygletWindow = ObjCClass('PygletWindow')
PygletToolWindow = ObjCClass('PygletToolWindow')



import unittest

from ...base.data import PygletTestCase

try:
    from pyglet.image.codecs import gdkpixbuf2
except ImportError:
    gdkpixbuf2 = None


@unittest.skipIf(gdkpixbuf2 is None, 'GdkPixBuf not available')
class GdkPixBufTest(PygletTestCase):
    def test_load_image(self):
        filename = self.get_test_data_file('images', '8bpp.gif')
        with open(filename, 'rb') as f:
            loader = gdkpixbuf2.GdkPixBufLoader(filename, f)
            pixbuf = loader.get_pixbuf()
            assert pixbuf is not None
            assert pixbuf.height == 257
            assert pixbuf.width == 235
            assert pixbuf.channels == 4
            assert pixbuf.has_alpha
            image = pixbuf.to_image()
            assert image is not None
            assert image.height == 257
            assert image.width == 235

    def test_load_image_requires_loader_close(self):
        """Some image files require the loader to be closed to force it
        to read and parse all data."""
        filename = self.get_test_data_file('images', 'gdk_close.png')
        with open(filename, 'rb') as f:
            loader = gdkpixbuf2.GdkPixBufLoader(filename, f)
            pixbuf = loader.get_pixbuf()
            assert pixbuf is not None
            assert pixbuf.height == 200
            assert pixbuf.width == 200
            assert pixbuf.channels == 4
            assert pixbuf.has_alpha
            image = pixbuf.to_image()
            assert image is not None
            assert image.height == 200
            assert image.width == 200

    def test_load_animation(self):
        filename = self.get_test_data_file('images', 'dinosaur.gif')
        with open(filename, 'rb') as f:
            loader = gdkpixbuf2.GdkPixBufLoader(filename, f)
            gdk_anim = loader.get_animation()
            assert gdk_anim is not None
            anim = gdk_anim.to_animation()
            assert anim is not None
            assert len(anim.frames) == 12
            for frame in anim.frames:
                assert frame.image is not None
                assert frame.duration is not None
                assert frame.duration == 0.1
                assert frame.image.width == 129
                assert frame.image.height == 79

    def test_incomplete_load(self):
        """Start loading a file and stop. Should clean up correctly."""
        filename = self.get_test_data_file('images', 'gdk_close.png')
        with open(filename, 'rb') as f:
            loader = gdkpixbuf2.GdkPixBufLoader(filename, f)
            del loader



"""
Test pyglet font package
"""

import pyglet


def test_load_privatefont(test_data):
    file = test_data.get_file('fonts', 'action_man.ttf')
    pyglet.font.add_file(file)
    myfont = pyglet.font.load("Action Man", size=12, dpi=96)
    assert myfont.name == "Action Man"


def test_load_privatefont_from_list(test_data):
    file = test_data.get_file('fonts', 'action_man.ttf')
    pyglet.font.add_file(file)
    # First font in the list should be returned:
    myfont = pyglet.font.load(["Action Man", "Arial"], size=12, dpi=96)
    assert myfont.name == "Action Man"



#!/usr/bin/env python
from setuptools import setup, find_packages


# Parse version number from pyglet/__init__.py:
with open('pyglet/__init__.py') as f:
    info = {}
    for line in f:
        if line.startswith('version'):
            exec(line, info)
            break


setup_info = dict(
    name='pyglet',
    version=info['version'],
    author='Alex Holkner',
    author_email='Alex.Holkner@gmail.com',
    url='http://pyglet.readthedocs.org/en/latest/',
    download_url='http://pypi.python.org/pypi/pyglet',
    project_urls={
        'Documentation': 'https://pyglet.readthedocs.io/en/latest',
        'Source': 'https://github.com/pyglet/pyglet',
        'Tracker': 'https://github.com/pyglet/pyglet/issues',
    },
    description='Cross-platform windowing and multimedia library',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Games/Entertainment',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    # Package info
    packages=['pyglet'] + ['pyglet.' + pkg for pkg in find_packages('pyglet')],

    # Add _ prefix to the names of temporary build dirs
    options={'build': {'build_base': '_build'}, },
    zip_safe=True,
)

setup(**setup_info)



from __future__ import annotations

from ctypes import byref

from pyglet import gl
from pyglet.display.headless import HeadlessCanvas
from pyglet.libs.egl import egl


from .base import DisplayConfig, Config, Context

_fake_gl_attributes = {
    'double_buffer': 0,
    'stereo': 0,
    'aux_buffers': 0,
    'accum_red_size': 0,
    'accum_green_size': 0,
    'accum_blue_size': 0,
    'accum_alpha_size': 0,
}


class HeadlessConfig(Config):  # noqa: D101
    def match(self, canvas: HeadlessCanvas) -> list[HeadlessDisplayConfig]:
        if not isinstance(canvas, HeadlessCanvas):
            msg = 'Canvas must be an instance of HeadlessCanvas'
            raise RuntimeError(msg)

        display_connection = canvas.display._display_connection  # noqa: SLF001

        # Construct array of attributes
        attrs = []
        for name, value in self.get_gl_attributes():
            if name == 'double_buffer':
                continue
            attr = HeadlessDisplayConfig.attribute_ids.get(name, None)
            if attr and value is not None:
                attrs.extend([attr, int(value)])
        attrs.extend([egl.EGL_SURFACE_TYPE, egl.EGL_PBUFFER_BIT])
        if self.opengl_api == "gl":
            attrs.extend([egl.EGL_RENDERABLE_TYPE, egl.EGL_OPENGL_BIT])
        elif self.opengl_api == "gles":
            attrs.extend([egl.EGL_RENDERABLE_TYPE, egl.EGL_OPENGL_ES3_BIT])
        else:
            msg = f"Unknown OpenGL API: {self.opengl_api}"
            raise ValueError(msg)
        attrs.extend([egl.EGL_NONE])
        attrs_list = (egl.EGLint * len(attrs))(*attrs)

        num_config = egl.EGLint()
        egl.eglChooseConfig(display_connection, attrs_list, None, 0, byref(num_config))
        configs = (egl.EGLConfig * num_config.value)()
        egl.eglChooseConfig(display_connection, attrs_list, configs,
                            num_config.value, byref(num_config))

        return [HeadlessDisplayConfig(canvas, c, self) for c in configs]


class HeadlessDisplayConfig(DisplayConfig):  # noqa: D101
    canvas: HeadlessCanvas
    config: HeadlessConfig

    attribute_ids = {  # noqa: RUF012
        'buffer_size': egl.EGL_BUFFER_SIZE,
        'level': egl.EGL_LEVEL,  # Not supported
        'red_size': egl.EGL_RED_SIZE,
        'green_size': egl.EGL_GREEN_SIZE,
        'blue_size': egl.EGL_BLUE_SIZE,
        'alpha_size': egl.EGL_ALPHA_SIZE,
        'depth_size': egl.EGL_DEPTH_SIZE,
        'stencil_size': egl.EGL_STENCIL_SIZE,
        'sample_buffers': egl.EGL_SAMPLE_BUFFERS,
        'samples': egl.EGL_SAMPLES,
    }

    def __init__(self, canvas: HeadlessCanvas, egl_config: egl.EGLConfig, config: HeadlessConfig) -> None:  # noqa: D107
        super().__init__(canvas, config)
        self._egl_config = egl_config
        context_attribs = (egl.EGL_CONTEXT_MAJOR_VERSION, config.major_version or 2,
                           egl.EGL_CONTEXT_MINOR_VERSION, config.minor_version or 0,
                           egl.EGL_CONTEXT_OPENGL_FORWARD_COMPATIBLE, config.forward_compatible or 0,
                           egl.EGL_CONTEXT_OPENGL_DEBUG, config.debug or 0,
                           egl.EGL_NONE)
        self._context_attrib_array = (egl.EGLint * len(context_attribs))(*context_attribs)

        for name, attr in self.attribute_ids.items():
            value = egl.EGLint()
            egl.eglGetConfigAttrib(canvas.display._display_connection, egl_config, attr, byref(value))  # noqa: SLF001
            setattr(self, name, value.value)

        for name, value in _fake_gl_attributes.items():
            setattr(self, name, value)

    def compatible(self, canvas: HeadlessCanvas) -> bool:
        return isinstance(canvas, HeadlessCanvas)

    def create_context(self, share: HeadlessContext | None) -> HeadlessContext:
        return HeadlessContext(self, share)


class HeadlessContext(Context):  # noqa: D101
    display_connection: egl.EGLDisplay
    config: HeadlessDisplayConfig

    def __init__(self, config: HeadlessDisplayConfig, share: HeadlessContext | None) -> None:  # noqa: D107
        super().__init__(config, share)

        self.display_connection = config.canvas.display._display_connection  # noqa: SLF001

        self.egl_context = self._create_egl_context(share)
        if not self.egl_context:
            msg = 'Could not create GL context'
            raise gl.ContextException(msg)

    def _create_egl_context(self, share: HeadlessContext | None) -> egl.EGLContext:
        if share:
            share_context = share.egl_context
        else:
            share_context = None

        if self.config.opengl_api == "gl":
            egl.eglBindAPI(egl.EGL_OPENGL_API)
        elif self.config.opengl_api == "gles":
            egl.eglBindAPI(egl.EGL_OPENGL_ES_API)
        return egl.eglCreateContext(self.config.canvas.display._display_connection,  # noqa: SLF001
                                    self.config._egl_config, share_context,  # noqa: SLF001
                                    self.config._context_attrib_array)  # noqa: SLF001

    def attach(self, canvas: HeadlessCanvas) -> None:
        if canvas is self.canvas:
            return

        super().attach(canvas)

        self.egl_surface = canvas.egl_surface
        self.set_current()

    def set_current(self) -> None:
        egl.eglMakeCurrent(
            self.display_connection, self.egl_surface, self.egl_surface, self.egl_context)
        super().set_current()

    def detach(self) -> None:
        if not self.canvas:
            return

        self.set_current()
        gl.glFlush()  # needs to be in try/except?

        super().detach()

        egl.eglMakeCurrent(
            self.display_connection, 0, 0, None)
        self.egl_surface = None

    def destroy(self) -> None:
        super().destroy()
        if self.egl_context:
            egl.eglDestroyContext(self.display_connection, self.egl_context)
            self.egl_context = None

    def flip(self) -> None:
        if not self.egl_surface:
            return

        egl.eglSwapBuffers(self.display_connection, self.egl_surface)



from collections import namedtuple

CustomField = namedtuple("CustomField", "fields removals repositions")
Reposition = namedtuple("Reposition", "field after")

# Versions of the loaded libraries
versions = {
    'avcodec': 0,
    'avformat': 0,
    'avutil': 0,
    'swresample': 0,
    'swscale': 0,
}

# Group codecs by version they are usually packaged with.
release_versions = {
    4: {'avcodec': 58, 'avformat': 58, 'avutil': 56, 'swresample': 3, 'swscale': 5},  # 4.x
    5: {'avcodec': 59, 'avformat': 59, 'avutil': 57, 'swresample': 4, 'swscale': 6},  # 5.x
    6: {'avcodec': 60, 'avformat': 60, 'avutil': 58, 'swresample': 4, 'swscale': 7},  # 6.x
    7: {'avcodec': 61, 'avformat': 61, 'avutil': 59, 'swresample': 5, 'swscale': 8},  # 7.x
}

# Removals done per library and version.
_version_changes = {
    'avcodec': {},
    'avformat': {},
    'avutil': {},
    'swresample': {},
    'swscale': {}
}


def set_version(library, version):
    versions[library] = version


def add_version_changes(library, version, structure, fields, removals, repositions=None):
    if version not in _version_changes[library]:
        _version_changes[library][version] = {}

    if structure in _version_changes[library][version]:
        raise Exception("Structure: {} from: {} has already been added for version {}.".format(structure,
                                                                                               library,
                                                                                               version)
                        )

    _version_changes[library][version][structure] = CustomField(fields, removals, repositions)


def apply_version_changes():
    """Apply version changes to Structures in FFmpeg libraries.
       Field data can vary from version to version, however assigning _fields_ automatically assigns memory.
       _fields_ can also not be re-assigned. Use a temporary list that can be manipulated before setting the
       _fields_ of the Structure."""

    for library, data in _version_changes.items():
        for version in data:
            for structure, cf_data in _version_changes[library][version].items():
                if versions[library] == version:
                    if cf_data.removals:
                        for remove_field in cf_data.removals:
                            for field in list(cf_data.fields):
                                if field[0] == remove_field:
                                    cf_data.fields.remove(field)

                    if cf_data.repositions:
                        for reposition in cf_data.repositions:
                            data = None
                            insertId = None
                            for idx, field in enumerate(list(cf_data.fields)):
                                if field[0] == reposition.field:
                                    data = field
                                elif field[0] == reposition.after:
                                    insertId = idx

                            if data and insertId:
                                cf_data.fields.remove(data)
                                cf_data.fields.insert(insertId+1, data)
                            else:
                                print(f"Warning: {reposition} for {library} was not able to be processed.")

                    structure._fields_ = cf_data.fields



import math
import ctypes

from . import interface
from pyglet.media.drivers.base import AbstractAudioDriver, AbstractAudioPlayer, MediaEvent
from pyglet.media.drivers.listener import AbstractListener
from pyglet.media.player_worker_thread import PlayerWorkerThread
from pyglet.util import debug_print

_debug = debug_print('debug_media')


def _convert_coordinates(coordinates):
    x, y, z = coordinates
    return x, y, -z


def _gain2db(gain):
    """
    Convert linear gain in range [0.0, 1.0] to 100ths of dB.

    Power gain = P1/P2
    dB = 2 log(P1/P2)
    dB * 100 = 1000 * log(power gain)
    """
    if gain <= 0:
        return -10000
    return max(-10000, min(int(1000 * math.log2(min(gain, 1))), 0))


def _db2gain(db):
    """Convert 100ths of dB to linear gain."""
    return math.pow(10.0, float(db)/1000.0)


class DirectSoundDriver(AbstractAudioDriver):
    def __init__(self):
        self._ds_driver = interface.DirectSoundDriver()
        self._ds_listener = self._ds_driver.create_listener()
        self._listener = DirectSoundListener(self._ds_listener, self._ds_driver.primary_buffer)

        assert self._ds_driver is not None
        assert self._ds_listener is not None

        self.worker = PlayerWorkerThread()
        self.worker.start()

    def create_audio_player(self, source, player):
        assert self._ds_driver is not None
        return DirectSoundAudioPlayer(self, source, player)

    def get_listener(self):
        assert self._ds_driver is not None
        assert self._ds_listener is not None
        return self._listener

    def delete(self):
        if self._ds_listener is None:
            assert _debug("DirectSoundDriver.delete() on deleted driver, ignoring")
            return

        assert _debug("Deleting DirectSoundDriver")
        self.worker.stop()
        # Destroy listener before destroying driver
        self._ds_listener.delete()
        self._ds_listener = None
        self._ds_driver.delete()
        self._ds_driver = None
        assert _debug("DirectSoundDriver deleted.")


class DirectSoundListener(AbstractListener):
    def __init__(self, ds_listener, ds_primary_buffer):
        self._ds_listener = ds_listener
        self._ds_primary_buffer = ds_primary_buffer

    def _set_volume(self, volume):
        self._volume = volume
        self._ds_primary_buffer.volume = _gain2db(volume)

    def _set_position(self, position):
        self._position = position
        self._ds_listener.position = _convert_coordinates(position)

    def _set_forward_orientation(self, orientation):
        self._forward_orientation = orientation
        self._set_orientation()

    def _set_up_orientation(self, orientation):
        self._up_orientation = orientation
        self._set_orientation()

    def _set_orientation(self):
        self._ds_listener.orientation = (_convert_coordinates(self._forward_orientation)
                                         + _convert_coordinates(self._up_orientation))


class DirectSoundAudioPlayer(AbstractAudioPlayer):
    def __init__(self, driver, source, player):
        super().__init__(source, player)

        # We keep here a strong reference because the AudioDriver is anyway
        # a singleton object which will only be deleted when the application
        # shuts down. The AudioDriver does not keep a ref to the AudioPlayer.
        self.driver = driver

        # Desired play state. As the DS Buffer is just a circular buffer, it is
        # not possible to have it stop at an exact position if the source
        # should lag behind and cause it to underrun.
        self._playing = False

        # Need to cache these because pyglet API allows update separately, but
        # DSound requires both to be set at once.
        self._cone_inner_angle = 360
        self._cone_outer_angle = 360

        # Actual cursors for the circular DSBuffer. Will never exceed
        # `self._buffer_size`.
        self._play_cursor_ring = 0
        self._write_cursor_ring = 0

        # Theoretical write and play cursors for an infinite buffer.  play
        # cursor is always <= write cursor (when equal, underrun is
        # happening).
        self._write_cursor = 0
        self._play_cursor = 0

        # Cursor position of where the source ran out.
        # We are done once the play cursor crosses it.
        self._eos_cursor = None

        # Always set to just after the most recently written audio data; so
        # either undefined data or silence. The eos cursor is set to it once
        # the source is confirmed to have run out.
        self._possible_eos_cursor = 0

        # Whether the source has hit its end; protect against duplicate
        # dispatch of on_eos events.
        self._has_underrun = False

        # DSound buffer
        self._buffer_size = self._buffered_data_ideal_size
        self._ds_buffer = self.driver._ds_driver.create_buffer(source.audio_format, self._buffer_size)
        self._ds_buffer.current_position = 0

        assert self._ds_buffer.buffer_size == self._buffer_size

    def delete(self):
        if self.driver._ds_driver is not None:
            self.driver.worker.remove(self)
            self._ds_buffer.delete()

    def play(self):
        assert _debug('DirectSound play')

        if not self._playing:
            self._playing = True
            self._ds_buffer.play()

        self.driver.worker.add(self)
        assert _debug('return DirectSound play')

    def stop(self):
        assert _debug('DirectSound stop')
        self.driver.worker.remove(self)

        if self._playing:
            self._playing = False
            self._ds_buffer.stop()

        assert _debug('return DirectSound stop')

    def clear(self):
        assert _debug('DirectSound clear')
        super().clear()
        self._ds_buffer.current_position = 0
        self._play_cursor_ring = self._write_cursor_ring = 0
        self._play_cursor = self._write_cursor = 0
        self._eos_cursor = None
        self._possible_eos_cursor = 0
        self._has_underrun = False

    def get_play_cursor(self):
        return self._play_cursor

    def work(self):
        assert self._playing

        self._update_play_cursor()
        self.dispatch_media_events(self._play_cursor)

        if self._eos_cursor is None:
            self._maybe_fill()
            return

        # Source exhausted, check whether we hit the end
        if not self._has_underrun and self._play_cursor > self._eos_cursor:
            self._has_underrun = True
            assert _debug('DirectSoundAudioPlayer: Dispatching eos')
            MediaEvent('on_eos').sync_dispatch_to_player(self.player)

        # While we are still playing / waiting for the on_eos to be dispatched for
        # the player to stop, the buffer continues playing. Ensure that silence is
        # filled.
        if (used := self._get_used_buffer_space()) < self._buffered_data_comfortable_limit:
            self._write(None, self._buffer_size - used)

    def _maybe_fill(self):
        if (used := self._get_used_buffer_space()) < self._buffered_data_comfortable_limit:
            self._refill(self.source.audio_format.align(self._buffer_size - used))

    def _refill(self, size):
        """Refill the next `size` bytes in the buffer using the source.
        `size` must be aligned.
        """
        audio_data = self._get_and_compensate_audio_data(size, self._play_cursor)
        if audio_data is None:
            assert _debug('DirectSoundAudioPlayer: Out of audio data')
            if self._eos_cursor is None:
                self._eos_cursor = self._possible_eos_cursor
            self._write(None, size)

        else:
            assert _debug(f'DirectSoundAudioPlayer: Got {audio_data.length} bytes of audio data')
            self.append_events(self._write_cursor, audio_data.events)
            self._write(audio_data, size)

    def _update_play_cursor(self):
        play_cursor_ring = self._ds_buffer.current_position.play_cursor
        if play_cursor_ring < self._play_cursor_ring:
            # Wrapped around
            self._play_cursor += self._buffer_size - self._play_cursor_ring
            self._play_cursor += play_cursor_ring
        else:
            self._play_cursor += play_cursor_ring - self._play_cursor_ring
        self._play_cursor_ring = play_cursor_ring

    def _get_used_buffer_space(self):
        return max(self._write_cursor - self._play_cursor, 0)

    def _write(self, audio_data, region_size):
        """Write data into the circular DSBuffer, starting at _write_cursor_ring.
        May supply None as audio_data to only write silence.
        If the audio data is not sufficient, will fill silence afterwards.
        If too much audio data is supplied, will write as much as fits.
        """
        if region_size == 0:
            return

        if audio_data is None:
            audio_size = 0
        else:
            audio_size = min(region_size, audio_data.length)
            self._possible_eos_cursor = self._write_cursor + audio_size
            audio_ptr = audio_data.pointer

        assert _debug(f'Writing {region_size}B ({audio_size}B data, {region_size - audio_size}B silence)')

        write_ptr = self._ds_buffer.lock(self._write_cursor_ring, region_size)

        a1_size = write_ptr.audio_length_1.value
        a2_size = write_ptr.audio_length_2.value

        assert 0 < region_size <= self._buffer_size
        assert region_size == a1_size + a2_size

        a2_silence = a2_size
        s = 0x80 if self.source.audio_format.sample_size == 8 else 0

        if audio_size < a1_size:
            if audio_size > 0:
                ctypes.memmove(write_ptr.audio_ptr_1, audio_ptr, audio_size)
            ctypes.memset(write_ptr.audio_ptr_1.value + audio_size, s, a1_size - audio_size)
        else:
            if a1_size > 0:
                ctypes.memmove(write_ptr.audio_ptr_1, audio_ptr, a1_size)
            if write_ptr.audio_ptr_2 and (a2_audio := (audio_size - a1_size)) > 0:
                ctypes.memmove(write_ptr.audio_ptr_2, audio_ptr + a1_size, a2_audio)
                a2_silence -= a2_audio
        if write_ptr.audio_ptr_2 and a2_silence > 0:
            ctypes.memset(write_ptr.audio_ptr_2.value + (a2_size - a2_silence), s, a2_silence)

        self._ds_buffer.unlock(write_ptr)

        self._write_cursor += region_size
        self._write_cursor_ring += region_size
        self._write_cursor_ring %= self._buffer_size

    def set_volume(self, volume):
        self._ds_buffer.volume = _gain2db(volume)

    def set_position(self, position):
        if self._ds_buffer.is3d:
            self._ds_buffer.position = _convert_coordinates(position)

    def set_min_distance(self, min_distance):
        if self._ds_buffer.is3d:
            self._ds_buffer.min_distance = min_distance

    def set_max_distance(self, max_distance):
        if self._ds_buffer.is3d:
            self._ds_buffer.max_distance = max_distance

    def set_pitch(self, pitch):
        frequency = int(pitch * self.source.audio_format.sample_rate)
        self._ds_buffer.frequency = frequency

    def set_cone_orientation(self, cone_orientation):
        if self._ds_buffer.is3d:
            self._ds_buffer.cone_orientation = _convert_coordinates(cone_orientation)

    def set_cone_inner_angle(self, cone_inner_angle):
        if self._ds_buffer.is3d:
            self._cone_inner_angle = int(cone_inner_angle)
            self._set_cone_angles()

    def set_cone_outer_angle(self, cone_outer_angle):
        if self._ds_buffer.is3d:
            self._cone_outer_angle = int(cone_outer_angle)
            self._set_cone_angles()

    def _set_cone_angles(self):
        inner = min(self._cone_inner_angle, self._cone_outer_angle)
        outer = max(self._cone_inner_angle, self._cone_outer_angle)
        self._ds_buffer.set_cone_angles(inner, outer)

    def set_cone_outer_gain(self, cone_outer_gain):
        if self._ds_buffer.is3d:
            volume = _gain2db(cone_outer_gain)
            self._ds_buffer.cone_outside_volume = volume

    def prefill_audio(self):
        self._maybe_fill()



from __future__ import annotations

from typing import Any, Callable, Sequence

import pyglet.lib
from pyglet.gl.lib import decorate_function, missing_function

from typing import Callable

__all__ = ['link_GL', 'link_AGL']

gl_lib = pyglet.lib.load_library(framework='OpenGL')
agl_lib = pyglet.lib.load_library(framework='AGL')


def link_GL(name: str, restype: Any, argtypes: Any, requires: str | None = None,  # noqa: N802, D103
            suggestions: Sequence[str] | None = None) -> Callable[..., Any]:
    try:
        func = getattr(gl_lib, name)
        func.restype = restype
        func.argtypes = argtypes
        decorate_function(func, name)
        return func
    except AttributeError:
        return missing_function(name, requires, suggestions)


def link_AGL(name: str, restype: Any, argtypes: Any, requires: str | None = None,  # noqa: N802, D103
            suggestions: Sequence[str] | None = None) -> Callable[..., Any]:
    try:
        func = getattr(agl_lib, name)
        func.restype = restype
        func.argtypes = argtypes
        decorate_function(func, name)
        return func
    except AttributeError:
        return missing_function(name, requires, suggestions)



import unittest
from unittest.mock import patch, MagicMock

GL_TRIANGLES = 4

# The default shader attributes.
# Update this if you change the default shader.
default_shader_attributes = {
    'colors': {'type': 35666, 'size': 1, 'location': 1, 'count': 4, 'format': 'f', 'instance': False},
    'position': {'type': 35665, 'size': 1, 'location': 0, 'count': 3, 'format': 'f', 'instance': False},
    'tex_coords': {'type': 35665, 'size': 1, 'location': 2, 'count': 3, 'format': 'f', 'instance': False},
}


class TestDefaultShader(unittest.TestCase):

    def setUp(self):
        """Setup the fake ShaderProgram for all of the tests."""
        patcher = patch.multiple(
            'pyglet.graphics.shader',
            _link_program=patch('pyglet.graphics.shader._link_program', return_value=1).start(),
            _introspect_attributes=patch('pyglet.graphics.shader._introspect_attributes',
                                         return_value=default_shader_attributes).start(),
            _introspect_uniforms=patch('pyglet.graphics.shader._introspect_uniforms',
                                       return_value={}).start(),
            _introspect_uniform_blocks=patch('pyglet.graphics.shader._introspect_uniform_blocks',
                                             return_value={}).start(),
            glEnableVertexAttribArray=patch('pyglet.graphics.shader.glEnableVertexAttribArray', new_callable=MagicMock()).start(),
            glVertexAttribPointer=patch('pyglet.graphics.shader.glVertexAttribPointer',
                                            new_callable=MagicMock()).start(),
        )
        self.addCleanup(patcher.stop)

        patcher_gl = patch.multiple(
            'pyglet.gl',
            current_context=patch('pyglet.gl.current_context', create=True).start(),
            glUseProgram=patch('pyglet.gl.glUseProgram').start(),
            glDeleteProgram=patch('pyglet.gl.glDeleteProgram').start(),
            glGenVertexArrays=patch('pyglet.gl.glGenVertexArrays').start(),
        )
        self.addCleanup(patcher_gl.stop)
        
        # These are imported already before this test runs, so we need to override them. Imports fixed.
        patcher_vertexbuffer = patch.multiple(
            'pyglet.graphics.vertexbuffer',
            glBindBuffer=MagicMock(),
            glBufferData=MagicMock(),
            glBufferSubData=MagicMock(),
            glGenBuffers=MagicMock(),
        )
        self.addCleanup(patcher_vertexbuffer.stop)
        patcher_vertexbuffer.start()
        
        patcher_vertexarray = patch.multiple(
            'pyglet.graphics.vertexarray',
            glGenVertexArrays=MagicMock(),
            glBindVertexArray=MagicMock(),
        )
        self.addCleanup(patcher_vertexarray.stop)
        patcher_vertexarray.start()
        
        # Create mock shaders
        mock_shader1 = MagicMock()
        mock_shader2 = MagicMock()

        from pyglet.graphics import Batch
        from pyglet.graphics.shader import ShaderProgram

        # Patch default batch.
        patch('pyglet.graphics.get_default_batch', return_value=Batch()).start()

        self.program = ShaderProgram(mock_shader1, mock_shader2)

        self.batch = Batch()

    def test_index_vertex_list_create_no_batch_no_group(self):
        vlist = self.program._vertex_list_create(  # noqa: SLF001
            count=4,
            mode=GL_TRIANGLES,
            indices=[0, 1, 2, 0, 2, 3],
            instances=None,
            batch=None,
            group=None,
            position=('f', (400, 400, 0, 400 + 50, 400, 0, 400 + 50, 400 + 50, 0, 400, 400 + 50, 0)),
            colors=('f', (1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1)),
        )

        assert vlist is not None
        assert vlist.count == 4

    def test_index_vertex_list_create_batch_no_group(self):
        vlist = self.program._vertex_list_create(  # noqa: SLF001
            count=4,
            mode=GL_TRIANGLES,
            indices=[0, 1, 2, 0, 2, 3],
            instances=None,
            batch=self.batch,
            group=None,
            position=('f', (400, 400, 0, 400 + 50, 400, 0, 400 + 50, 400 + 50, 0, 400, 400 + 50, 0)),
            colors=('f', (1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1)),
        )

        assert vlist is not None
        assert vlist.count == 4
        
    def test_index_vertex_list_create_batch_group(self):
        from pyglet.graphics import Group
        vlist = self.program._vertex_list_create(  # noqa: SLF001
            count=4,
            mode=GL_TRIANGLES,
            indices=[0, 1, 2, 0, 2, 3],
            instances=None,
            batch=self.batch,
            group=Group(),
            position=('f', (400, 400, 0, 400 + 50, 400, 0, 400 + 50, 400 + 50, 0, 400, 400 + 50, 0)),
            colors=('f', (1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1)),
        )

        assert vlist is not None
        assert vlist.count == 4
        
    def test_index_vertex_list_create_no_batch_group(self):
        from pyglet.graphics import Group
        vlist = self.program._vertex_list_create(  # noqa: SLF001
            count=4,
            mode=GL_TRIANGLES,
            indices=[0, 1, 2, 0, 2, 3],
            instances=None,
            batch=None,
            group=Group(),
            position=('f', (400, 400, 0, 400 + 50, 400, 0, 400 + 50, 400 + 50, 0, 400, 400 + 50, 0)),
            colors=('f', (1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1, 1, 0.5, 0.2, 1)),
        )

        assert vlist is not None
        assert vlist.count == 4
        
if __name__ == '__main__':
    unittest.main()
